"""
The wntr.metrics.topographic module contains topographic metrics that are not
available directly with NetworkX.  Functions in this module operate on a 
NetworkX MultiDiGraph, which can be created by calling ``G = wn.to_graph()``
"""
import networkx as nx
import numpy as np
import pandas as pd
import logging
import warnings

logger = logging.getLogger(__name__)

def terminal_nodes(G):
    """
    Nodes with degree 1

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph

    Returns
    -------
    List of terminal nodes
    
    """
    node_degree = dict(G.degree())
    terminal_nodes = [k for k,v in node_degree.items() if v == 1]

    return terminal_nodes

def bridges(G):
    """
    Bridge links

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph

    Returns
    -------
    List of links that are bridges
    
    """
    uG = G.to_undirected() # uses an undirected graph
    bridge_links = []
    bridges = nx.bridges(nx.Graph(uG)) # not implemented for multigraph
    for br in bridges:
        for key in uG[br[0]][br[1]].keys():
            bridge_links.append(key)
        
    return bridge_links

def central_point_dominance(G):
    """
    Central point dominance
    
    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Central point dominance (float)
    
    """
    uG = G.to_undirected() # uses an undirected graph
    bet_cen = nx.betweenness_centrality(nx.Graph(uG)) # not implemented for multigraph 
    bet_cen = list(bet_cen.values())
    cpd = sum(max(bet_cen) - np.array(bet_cen))/(len(bet_cen)-1)

    return cpd

def spectral_gap(G):
    """
    Spectral gap
    
    Difference in the first and second eigenvalue of the adjacency matrix
    
    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Spectral gap (float)
    
    """
    uG = G.to_undirected() # uses an undirected graph
    eig = nx.adjacency_spectrum(uG)
    spectral_gap = abs(eig[0] - eig[1])

    return spectral_gap.real

def algebraic_connectivity(G):
    """
    Algebraic connectivity
    
    Second smallest eigenvalue of the normalized Laplacian matrix of a network

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Algebraic connectivity (float)
    
    """
    uG = G.to_undirected() # uses an undirected graph
    eig = nx.laplacian_spectrum(uG)
    eig = np.sort(eig)
    alg_con = eig[1]

    return alg_con

def critical_ratio_defrag(G):
    """
    Critical ratio of defragmentation

    Parameters
    ----------
    G: networkx MultiDiGraph
        Graph
        
    Returns
    -------
    Critical ratio of defragmentation (float)
    
    """
    node_degree = dict(G.degree())
    tmp = np.mean(pow(np.array(list(node_degree.values())),2))
    fc = 1-(1/((tmp/np.mean(list(node_degree.values())))-1))

    return fc


def _links_in_simple_paths(G, sources, sinks):
    """
    Count all links in a simple path between sources and sinks

    Parameters
    -----------
    G: networkx MultiDiGraph
        Graph
    sources: list
        List of source nodes
    sinks: list
        List of sink nodes

    Returns
    -------
    Dictionary with the number of times each link is involved in a path
    
    """
    link_names = [name for (node1, node2, name) in list(G.edges(keys=True))]
    link_count = pd.Series(data = 0, index=link_names)

    for sink in sinks:
        for source in sources:
            if nx.has_path(G, source, sink):
                paths = nx.all_simple_paths(G,source,target=sink)
                for path in paths:
                    for i in range(len(path)-1):
                        links = list(G[path[i]][path[i+1]].keys())
                        for link in links:
                            link_count[link] = link_count[link]+1

    return link_count

def valve_segments(G, valve_layer):
    """
    Valve segmentation

    Parameters
    -----------
    G: networkx MultiDiGraph
        Graph
    valve_layer: pandas DataFrame
        Valve layer, defined by node and link pairs (for example, valve 0 is 
        on link A and protects node B). The valve_layer DataFrame is indexed by
        valve number, with columns named 'node' and 'link'.

    Returns
    -------
    node_segments: pandas Series
       Segment number for each node, indexed by node name
    link_segments: pandas Series
        Segment number for each link, indexed by link name
    segment_size: pandas DataFrame
        Number of nodes and links in each segment. The DataFrame is indexed by 
        segment number, with columns named 'node' and 'link'.
    """
    

    # First check for duplicate valves
    if valve_layer.duplicated().any():
        valve_layer.drop_duplicates(inplace = True)
        warnings.warn('One or more valves were duplicated in `valve_layer`; duplicates are ignored.', stacklevel=0)

    # Convert the graph to an undirected graph
    uG = G.to_undirected()

    # Node and link names
    # Append N_ and L_ to node and link names to account for naming overlap
    node_names = ['N_'+n for n in uG.nodes()]
    link_names = ['L_'+k for u,v,k in uG.edges(keys=True)]
    all_names = node_names + link_names

    # Initialization for labelling
    seg_index = 0
    seg_label = np.zeros(shape=(len(all_names)), dtype=int)

    # Find and label links isolated by valves, EG 0|----|0
    for start_node, end_node, link_name in uG.edges(keys=True):
        link_valves = valve_layer[valve_layer['link']==link_name]
        if set(link_valves['node']) >= set([start_node, end_node]):
            seg_index += 1
            seg_label[all_names.index('L_'+link_name)] = seg_index

    # Find and label nodes isolated by valves, EG 0----|0|----0
    for node_name in node_names:
        node_valves = valve_layer[valve_layer['node']==node_name]
        node_links = [k for u,v,k in uG.edges(node_name[2:], keys=True)]
        if set(node_valves['link']) >= set(node_links):
            seg_index += 1
            seg_label[all_names.index(node_name)] = seg_index
    
    # Collect valved link names
    valved_link_names = list(valve_layer['link'].unique()) 

    # Remove valved edges from G
    valved_edges = []
    for edge in uG.edges:
        link_name = edge[2]
        if link_name in valved_link_names:
            valved_edges.append(edge)
    uG.remove_edges_from(valved_edges)

    ## Label unvalved portion of graph using connected components

    # Assign labels to nodes
    for component in nx.connected_components(uG):
        # component is a set of node names
        seg_index += 1
        for node in component:
            index = all_names.index('N_'+node)
            seg_label[index] = seg_index

    # Assign labels to links based on labelling of their nodes
    for edge in uG.edges(keys=True):
        node1, node2, link_name = edge
        node1 = all_names.index('N_'+node1)
        link_index = all_names.index('L_'+link_name)
        seg_label[link_index] = seg_label[node1]

    ## Label valved portion of graph
    for valved_edge in valved_edges:
        node1_name, node2_name, link_name = valved_edge
        link_valves = valve_layer[valve_layer['link']==link_name]
        link_index = all_names.index('L_'+link_name)

        # When link only has one valved node, locate unvalved node
        # and label link and unvalved node together
        if link_valves.shape[0] == 1:
            both_node_names = [node1_name, node2_name]
            valved_node_name = link_valves.iloc[0]['node']
            both_node_names.remove(valved_node_name)
            unvalved_node_name = both_node_names[0]
            unvalved_node_index = all_names.index('N_'+unvalved_node_name)
            if seg_label[unvalved_node_index] == 0:
                seg_index += 1
                seg_label[unvalved_node_index] = seg_index
                seg_label[link_index] = seg_index
            else:
                seg_label[link_index] = seg_label[unvalved_node_index]

        # Links with link_valves.size == 2 are already labelled (isolated link)
        elif link_valves.shape[0] == 2:
            continue
        else:
            raise Exception("Each link should have a maximum of two valves.")


    # Finalize results
    seg_labels_index = all_names
    seg_label = pd.Series(seg_label, index=seg_labels_index, dtype=int)

    node_segments = seg_label[node_names]
    link_segments = seg_label[link_names]

    # Remove "N_" and "L_" from nodes and links
    node_segments.index = node_segments.index.str[2::]
    link_segments.index = link_segments.index.str[2::]

    # Extract segment sizes, for nodes and links
    seg_link_sizes = link_segments.value_counts().rename('link')
    seg_node_sizes = node_segments.value_counts().rename('node')
    seg_sizes = pd.concat([seg_link_sizes, seg_node_sizes], axis=1).fillna(0)
    seg_sizes = seg_sizes.astype(int)

    return node_segments, link_segments, seg_sizes


def valve_segment_attributes(valve_layer, node_segments, link_segments, 
                             demand=None, length=None):
    """
    Valve segment attributes include 1) the number of valves surrounding each valve
    and (optionally) the increase in segment demand if a given valve is removed and 
    the increase in segment pipe length if a given valve is removed. 
    
    The increase in segment demand is  expressed as a fraction of the 
    max segment demand associated with that valve.  Likewise, 
    the increase in segment pipe length is expressed as a fraction of the 
    max segment pipe length associated with that valve.
	
    Parameters
    ----------
    valve_layer: pandas DataFrame
        Valve layer, defined by node and link pairs (for example, valve 0 is 
        on link A and protects node B). The valve_layer DataFrame is indexed by
        valve number, with columns named 'node' and 'link'.
    
    node_segments: pandas Series
       Segment number for each node, indexed by node name.
       node_segments can be computed using `wntr.metrics.topographic.valve_segments`
       
    link_segments: pandas Series
        Segment number for each link, indexed by link name. 
        link_segments can be computed using `wntr.metrics.topographic.valve_segments`

    demands: pandas Series, optional
        Node demand, the average expected node demand can be computed using 
        wntr.metrics.average_expected_demand(wn). 
        Demand from simulation results can also be used.
        
    lengths: pandas Series, optional
        A list of 'length' attributes for each link in the network.
        The output from wn.query_link_attribute('length')

    Returns
    -------
    pandas DataFrame 
        Valve segement attributes, indexed by valve number, that contains:
    
       * num_surround: number of valves surrounding each valve
       * demand_increase: increase in segment demand if a given valve is removed, expressed as a fraction
       * length_increase: increase in segment pipe length if a given valve is removed, expressed as a fraction
    """
    valve_attr = pd.DataFrame()
    
    valve_attr['num_surround'] = _valve_criticality(valve_layer, node_segments, link_segments)
    
    if demand is not None:
        valve_attr['demand_increase'] = _valve_criticality_demand(demand, valve_layer, node_segments, link_segments)
    
    if length is not None:
        valve_attr['length_increase'] = _valve_criticality_length(length, valve_layer, node_segments, link_segments)
                                           
    return valve_attr

def _valve_criticality(valve_layer, node_segments, link_segments):
    """
	Returns the number of valves surrounding each valve
	
    """
    # Assess the number of valves in the system
    n_valves = len(valve_layer)   

    # Calculate valve-based valve criticality
    VC = {}
    for i in range(n_valves):
        # identify the node-side and link-side segments
        node_seg = node_segments[valve_layer.loc[i,'node']]
        link_seg = link_segments[valve_layer.loc[i,'link']] 
        # if the node and link are in the same segment, set criticality to 0
        if node_seg == link_seg:
            VC_val_i = 0 
        else:
            V_list = []
            # identify links and nodes in surrounding segments
            links_in_segs = link_segments[(link_segments == link_seg) | (link_segments == node_seg)].index
            nodes_in_segs = node_segments[(node_segments == link_seg) | (node_segments == node_seg)].index
            # add unique valves to the V_list from the link list
            for link in links_in_segs:
                valves = valve_layer[valve_layer['link'] == link].index
                if len(valves) == 0:
                    pass
                else:
                    for valve in valves:
                        if valve in V_list:
                            pass
                        else:
                            V_list.append(valve)
            # add unique valves to the V_list from the node list
            for node in nodes_in_segs:
                valves = valve_layer[valve_layer['node'] == node].index
                if len(valves) == 0:
                    pass
                else:
                    for valve in valves:
                        if valve in V_list:
                            pass
                        else:
                            V_list.append(valve)
            # calculate valve-based criticality for the valve
            # count the number of valves in the list, minus the valve in question
            VC_val_i = len(V_list) - 1

        VC[i] = VC_val_i
    
    VC = pd.Series(VC)
    
    return VC

def _valve_criticality_length(link_lengths, valve_layer, node_segments, link_segments):
    """
	Returns the ratio of the segment lengths on either side of the valve
    """
    # Assess the number of valves in the system
    n_valves = len(valve_layer)   

    # Calculate the length-based valve crticiality
    VC = {}
    
    for i in range(n_valves):
		# identify the node-side and link-side segments
        node_seg = node_segments[valve_layer.loc[i,'node']]
        link_seg = link_segments[valve_layer.loc[i,'link']]
        
        # if the node and link are in the same segment, set criticality to 0
        if node_seg == link_seg:
            VC_len_i = 0
        else:
            # calculate total length of links in the node segment
            links_in_node_seg = link_segments[link_segments == node_seg].index
            n_ixs = link_lengths.index.intersection(links_in_node_seg)
            L_node = link_lengths[n_ixs].sum()

            # calculate total length of links in the link segment
            links_in_link_seg = link_segments[link_segments == link_seg].index
            l_ixs = link_lengths.index.intersection(links_in_link_seg)
            L_link = link_lengths[l_ixs].sum()

            # calculate link length criticality for the valve
            if L_node == 0 and L_link == 0:
                VC_len_i = 0.0
            else:
                VC_len_i = (L_link + L_node) / max(L_link, L_node) - 1

        VC[i] = VC_len_i
    
    VC = pd.Series(VC)
    
    return VC

def _valve_criticality_demand(node_demands, valve_layer, node_segments, link_segments):
    """
	Returns the ratio of node demands on either side of a valve.
    """
    # Assess the number of valves in the system
    n_valves = len(valve_layer)         

    # Calculate the demand-based valve crticiality
    VC = {}

    for i in range(n_valves):
        # identify the node-side and link-side segments
        node_seg = node_segments[valve_layer.loc[i,'node']]
        link_seg = link_segments[valve_layer.loc[i,'link']] 
        # if the node and link are in the same segment, set criticality to 0
        if node_seg == link_seg:
            VC_dem_i = 0.0
        else:
            # calculate total demand in the node segment
            nodes_in_node_seg = node_segments[node_segments == node_seg].index
            n_ixs = node_demands.index.intersection(nodes_in_node_seg)
            D_node = node_demands.loc[n_ixs].sum()

            # calculate total demand in the link segment
            nodes_in_link_seg = node_segments[node_segments == link_seg].index
            l_ixs = node_demands.index.intersection(nodes_in_link_seg)
            D_link = node_demands[l_ixs].sum()

            # calculate demand criticality for the valve
            if D_node == 0 and D_link == 0:
                VC_dem_i = 0
            else:
                VC_dem_i = (D_link + D_node) / max(D_link, D_node) - 1

        VC[i] = VC_dem_i
    
    VC = pd.Series(VC)
    
    return VC




# coding: utf-8
"""
The wntr.epanet.msx.enums module contains EPANET-MSX enum types, for use in
toolkit API calls.
"""

from enum import IntEnum
from wntr.utils.enumtools import add_get

@add_get(prefix='MSX_')
class TkObjectType(IntEnum):
    r"""Enumeration for object type used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        NODE
        LINK
        PIPE
        TANK
        SPECIES
        TERM
        PARAMETER
        CONSTANT
        PATTERN
    """
    NODE = 0
    """EPANET node"""
    LINK = 1
    """EPANET link"""
    PIPE = 1
    """EPANET pipe"""
    TANK = 2
    """EPANET tank"""
    SPECIES = 3
    """MSX species"""
    TERM = 4
    """MSX term"""
    PARAMETER = 5
    """MSX parameter"""
    CONSTANT = 6
    """MSX constant"""
    PATTERN = 7
    """**MSX** pattern"""
    MAX_OBJECTS = 8


@add_get(prefix='MSX_')
class TkSourceType(IntEnum):
    r"""Enumeration for source type used in EPANET-MSX.

    .. warning:: These enum values start with -1.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        NOSOURCE
        CONCEN
        MASS
        SETPOINT
        FLOWPACED
    """
    NOSOURCE = -1
    """No source"""
    CONCEN = 0
    """Concentration based source"""
    MASS = 1
    """Constant mass source"""
    SETPOINT = 2
    """Setpoint source"""
    FLOWPACED = 3
    """Flow-paced source"""


@add_get(prefix='MSX_')
class TkUnitSystem(IntEnum):
    r"""Enumeration for the units system used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        US
        SI
    """
    US = 0
    """US units (ft, ft2, gal)"""
    SI = 1
    """SI units (m, m2, m3)"""


@add_get(prefix='MSX_')
class TkFlowUnits(IntEnum):
    r"""Enumeration for the flow units used in EPANET-MSX (determined from
    EPANET INP file read in with the toolkit).

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        CFS
        GPM
        MGD
        IMGD
        AFD
        LPS
        LPM
        MLD
        CMH
        CMD
    """
    CFS = 0
    """cubic feet per second"""
    GPM = 1
    """gallons (US) per minute"""
    MGD = 2
    """million gallons (US) per day"""
    IMGD = 3
    """million Imperial gallons per day"""
    AFD = 4
    """acre-feet (US) per day"""
    LPS = 5
    """liters per second"""
    LPM = 6
    """liters per minute"""
    MLD = 7
    """million liters per day"""
    CMH = 8
    """cubic meters per hour"""
    CMD = 9
    """cubic meters per day"""


@add_get(prefix='MSX_')
class TkMixType(IntEnum):
    r"""Enumeration for the mixing model used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        MIX1
        MIX2
        FIFO
        LIFO
    """
    MIX1 = 0
    """full mixing, 1 compartment"""
    MIX2 = 1
    """full mixing, 2 comparments"""
    FIFO = 2
    """first in, first out"""
    LIFO = 3
    """last in, first out"""


@add_get(prefix='MSX_')
class TkSpeciesType(IntEnum):
    r"""Enumeration for species type used in EPANET-MSX.
    
    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        BULK
        WALL
    """
    BULK = 0
    """bulk species"""
    WALL = 1
    """wall species"""


@add_get(prefix='MSX_')
class TkExpressionType(IntEnum):
    r"""Enumeration for the expression type used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        NO_EXPR
        RATE
        FORMULA
        EQUIL
    """
    NO_EXPR = 0
    """no expression defined"""
    RATE = 1
    """expression is a rate expression"""
    FORMULA = 2
    """expression is a formula expression"""
    EQUIL = 3
    """expression is an equilibrium expression"""


@add_get(prefix='MSX_')
class TkSolverType(IntEnum):
    r"""Enumeration for the solver type used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        EUL
        RK5
        ROS2
    """
    EUL = 0
    """Euler solver"""
    RK5 = 1
    """Runga-Kutta 5th order solver"""
    ROS2 = 2
    """Ros 2nd order solver"""


@add_get(prefix='MSX_')
class TkCouplingType(IntEnum):
    r"""Enumeration for the coupling type option used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        NO_COUPLING
        FULL_COUPLING
    """
    NO_COUPLING = 0
    """no coupling"""
    FULL_COUPLING = 1
    """full coupling"""


@add_get(prefix='MSX_')
class TkMassUnits(IntEnum):
    r"""Enumeration for mass units used in EPANET-MSX.
    
    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        MG
        UG
        MOLE
        MMOLE
    """
    MG = 0
    """milligrams"""
    UG = 1
    """micrograms"""
    MOLE = 2
    """mole"""
    MMOLE = 3
    """millimole"""


@add_get(prefix='MSX_')
class TkAreaUnits(IntEnum):
    r"""Enumeration for area units used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        FT2
        M2
        CM2
    """
    FT2 = 0
    """square feet"""
    M2 = 1
    """square meters"""
    CM2 = 2
    """square centimeters"""


@add_get(prefix='MSX_')
class TkRateUnits(IntEnum):
    r"""Enumeration for rate units used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        SECONDS
        MINUTES
        HOURS
        DAYS
    """
    SECONDS = 0
    """per second"""
    MINUTES = 1
    """per minute"""
    HOURS = 2
    """per hour"""
    DAYS = 3
    """per day"""


@add_get(prefix='MSX_')
class TkUnits(IntEnum):
    r"""Position for units used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        LENGTH_UNITS
        DIAM_UNITS
        AREA_UNITS
        VOL_UNITS
        FLOW_UNITS
        CONC_UNITS
        RATE_UNITS
    """
    LENGTH_UNITS = 0
    """the length unit index"""
    DIAM_UNITS = 1
    """the diameter unit index"""
    AREA_UNITS = 2
    """the area unit index"""
    VOL_UNITS = 3
    """the volume unit index"""
    FLOW_UNITS = 4
    """the flow unit index"""
    CONC_UNITS = 5
    """the concentration unit index"""
    RATE_UNITS = 6
    """the rate unit index"""
    MAX_UNIT_TYPES = 7


@add_get(prefix='MSX_')
class TkHydVar(IntEnum):
    r"""Enumeration for hydraulic variable used in EPANET-MSX.
    
    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        DIAMETER
        FLOW
        VELOCITY
        REYNOLDS
        SHEAR
        FRICTION
        AREAVOL
        ROUGHNESS
        LENGTH
    """
    DIAMETER = 1
    """pipe diameter"""
    FLOW = 2
    """pipe flow rate"""
    VELOCITY = 3
    """segment velocity"""
    REYNOLDS = 4
    """Reynolds number"""
    SHEAR = 5
    """shear velocity"""
    FRICTION = 6
    """friction factor"""
    AREAVOL = 7
    """area / volume ratio"""
    ROUGHNESS = 8
    """roughness number"""
    LENGTH = 9
    """pipe or segment length"""
    MAX_HYD_VARS = 10


@add_get(prefix='MSX_')
class TkTstat(IntEnum):
    r"""Enumeration used for time statistic in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        SERIES
        AVERAGE
        MINIMUM
        MAXIMUM
        RANGE
    """
    SERIES = 0
    """output a time series"""
    AVERAGE = 1
    """output the average"""
    MINIMUM = 2
    """output the minimum"""
    MAXIMUM = 3
    """output the maximum"""
    RANGE = 4
    """output the range (max - min)"""


@add_get(prefix='MSX_')
class TkOption(IntEnum):
    r"""Enumeration used for choosing an option in EPANET-MSX toolkit.
    
    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        AREA_UNITS_OPTION
        RATE_UNITS_OPTION
        SOLVER_OPTION
        COUPLING_OPTION
        TIMESTEP_OPTION
        RTOL_OPTION
        ATOL_OPTION
        COMPILER_OPTION
        MAXSEGMENT_OPTION
        PECLETNUMBER_OPTION
    """
    AREA_UNITS_OPTION = 0
    """area units"""
    RATE_UNITS_OPTION = 1
    """rate units"""
    SOLVER_OPTION = 2
    """solver"""
    COUPLING_OPTION = 3
    """coupling"""
    TIMESTEP_OPTION = 4
    """timestep size"""
    RTOL_OPTION = 5
    """relative tolerance (global)"""
    ATOL_OPTION = 6
    """absolute tolerance (global)"""
    COMPILER_OPTION =7
    """compiler option"""
    MAXSEGMENT_OPTION = 8
    """max segments"""
    PECLETNUMBER_OPTION = 9
    """peclet number"""


@add_get(prefix='MSX_')
class TkCompiler(IntEnum):
    r"""Enumeration used for specifying compiler options in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        NO_COMPILER
        VC
        GC
    """
    NO_COMPILER = 0
    """do not compile reaction dynamics"""
    VC = 1
    """use Visual C to compile reaction dynamics"""
    GC = 2
    """use Gnu C to compile reaction dynamics"""


@add_get(prefix='MSX_')
class TkFileMode(IntEnum):
    r"""Enumeration for file model used in EPANET-MSX.

    .. warning:: These enum values start with 0.

    .. rubric:: Enum Members

    The following enum names are defined, and, if using the :meth:`get` method,
    then they are case insensitive and can be optionally prefixed with "MSX\_".

    .. autosummary::
        SCRATCH_FILE
        SAVED_FILE
        USED_FILE
    """
    SCRATCH_FILE = 0
    """use a scratch file"""
    SAVED_FILE = 1
    """save the file"""
    USED_FILE = 2
    """use a saved file"""



# coding: utf-8
"""
The wntr.epanet.msx io module contains methods for reading/writing EPANET
MSX input and output files.
"""

import datetime
import logging
import sys
from typing import Union

import numpy as np
import pandas as pd
from wntr.msx.elements import Constant, Parameter, Species, Term

import wntr.network
from wntr.epanet.msx.exceptions import EpanetMsxException, MSXValueError
from wntr.epanet.util import ENcomment
from wntr.msx.base import VariableType, SpeciesType
from wntr.msx.model import MsxModel

sys_default_enc = sys.getdefaultencoding()

logger = logging.getLogger(__name__)

MAX_LINE = 1024

_INP_SECTIONS = [
    "[TITLE]",
    "[OPTIONS]",
    "[SPECIES]",
    "[COEFFICIENTS]",
    "[TERMS]",
    "[PIPES]",
    "[TANKS]",
    "[SOURCES]",
    "[QUALITY]",
    "[PARAMETERS]",
    "[DIFFUSIVITY]",
    "[PATTERNS]",
    "[REPORT]",
]


def _split_line(line):
    _vc = line.split(";", 1)
    _cmnt = None
    _vals = None
    if len(_vc) == 0:
        pass
    elif len(_vc) == 1:
        _vals = _vc[0].split()
    elif _vc[0] == "":
        _cmnt = _vc[1].strip()
    else:
        _vals = _vc[0].split()
        _cmnt = _vc[1].strip()
    return _vals, _cmnt


class MsxFile(object):
    """An EPANET-MSX input file reader.

    .. rubric:: Class Methods
    .. autosummary::
        :nosignatures:

        read
        write

    """

    def __init__(self):
        self.rxn: MsxModel = None
        self.sections = dict()
        for sec in _INP_SECTIONS:
            self.sections[sec] = []
        self.top_comments = []
        self.patterns = dict()

    @classmethod
    def read(cls, msx_filename: str, rxn_model: MsxModel = None):
        """
        Read an EPANET-MSX input file (.msx) and load data into a MsxModel.
        Only MSX 2.0 files are recognized.

        Parameters
        ----------
        msx_file : str
            Filename of the .msx file to read in
        rxn_model : MsxModel, optional
            Multi-species water quality model to put data into,
            by default None (new model)

        Returns
        -------
        MsxModel
            Multi-species water quality model with the new species, reactions
            and other options added
        """
        if rxn_model is None:
            rxn_model = MsxModel()
        obj = cls()
        obj.rxn = rxn_model
        # if not isinstance(msx_file, list):
        #     msx_file = [msx_file]
        rxn_model._orig_file = msx_filename

        obj.patterns = dict()
        obj.top_comments = []
        obj.sections = dict()
        for sec in _INP_SECTIONS:
            obj.sections[sec] = []

        def _read():
            section = None
            lnum = 0
            edata = {"fname": msx_filename}
            with open(msx_filename, "r", encoding=sys_default_enc) as f:
                for line in f:
                    lnum += 1
                    edata["lnum"] = lnum
                    line = line.strip()
                    nwords = len(line.split())
                    if len(line) == 0 or nwords == 0:
                        # Blank line
                        continue
                    elif line.startswith("["):
                        vals = line.split(None, 1)
                        sec = vals[0].upper()
                        # Add handlers to deal with extra 'S'es (or missing 'S'es) in INP file
                        if sec not in _INP_SECTIONS:
                            trsec = sec.replace("]", "S]")
                            if trsec in _INP_SECTIONS:
                                sec = trsec
                        if sec not in _INP_SECTIONS:
                            trsec = sec.replace("S]", "]")
                            if trsec in _INP_SECTIONS:
                                sec = trsec
                        edata["sec"] = sec
                        if sec in _INP_SECTIONS:
                            section = sec
                            # logger.info('%(fname)s:%(lnum)-6d %(sec)13s section found' % edata)
                            continue
                        # elif sec == "[END]":
                        #     # logger.info('%(fname)s:%(lnum)-6d %(sec)13s end of file found' % edata)
                        #     section = None
                        #     break
                        else:
                            logger.warning('%(fname)s:%(lnum)d: Invalid section "%(sec)s"' % edata)
                            raise EpanetMsxException(201, "at line {}:\n{}".format(lnum, line))
                    elif section is None and line.startswith(";"):
                        obj.top_comments.append(line[1:])
                        continue
                    elif section is None:
                        logger.debug("Found confusing line: %s", repr(line))
                        raise EpanetMsxException(201, "at line {}:\n{}".format(lnum, line))
                    # We have text, and we are in a section
                    obj.sections[section].append((lnum, line))

        try:
            _read()
            obj._read_title()
            obj._read_options()
            obj._read_species()
            obj._read_coefficients()
            obj._read_terms()
            obj._read_pipes()
            obj._read_tanks()
            obj._read_source_dict()
            obj._read_quality()
            obj._read_parameters()
            obj._read_diffusivity()
            obj._read_patterns()
            obj._read_report()
            return obj.rxn
        except Exception as e:
            raise EpanetMsxException(200) from e

    @classmethod
    def write(cls, filename: str, msx: MsxModel):
        """Write an MSX input file.

        Parameters
        ----------
        filename : str
            Filename to write
        rxn : MsxModel
            Multi-species water quality model
        """
        obj = cls()
        obj.rxn = msx
        with open(filename, "w") as fout:
            fout.write("; WNTR-reactions MSX file generated {}\n".format(datetime.datetime.now()))
            fout.write("\n")
            obj._write_title(fout)
            obj._write_options(fout)
            obj._write_species(fout)
            obj._write_coefficients(fout)
            obj._write_terms(fout)
            obj._write_pipes(fout)
            obj._write_tanks(fout)
            obj._write_diffusivity(fout)
            obj._write_parameters(fout)
            obj._write_patterns(fout)
            obj._write_report(fout)
            obj._write_quality(fout)
            obj._write_source_dict(fout)

    def _read_title(self):
        lines = []
        title = None
        comments = ""
        for lnum, line in self.sections["[TITLE]"]:
            vals, comment = _split_line(line)
            if title is None and vals is not None:
                title = " ".join(vals)
            if comment:
                lines.append(comment.strip())
        if self.top_comments:
            comments = "\n".join(self.top_comments)
        if len(lines) > 0:
            comments = comments + ("\n" if comments else "") + "\n".join(lines)
        self.rxn.title = title
        self.rxn.description = comments

    def _read_options(self):
        for lnum, line in self.sections["[OPTIONS]"]:
            vals, comment = _split_line(line)
            try:
                if len(vals) < 2:
                    raise EpanetMsxException(402, "at line {} of [OPTIONS] section:\n{}".format(lnum, line))
                name, val = vals[0].upper(), vals[1].upper()
                if name == "AREA_UNITS":
                    self.rxn._options.area_units = val
                elif name == "RATE_UNITS":
                    self.rxn._options.rate_units = val
                elif name == "SOLVER":
                    self.rxn._options.solver = val
                elif name == "COUPLING":
                    self.rxn._options.coupling = val
                elif name == "TIMESTEP":
                    self.rxn._options.timestep = int(val)
                elif name == "ATOL":
                    self.rxn._options.atol = float(val)
                elif name == "RTOL":
                    self.rxn._options.rtol = float(val)
                elif name == "COMPILER":
                    self.rxn._options.compiler = val
                elif name == "SEGMENTS":
                    self.rxn._options.segments = int(val)
                elif name == "PECLET":
                    self.rxn._options.peclet = int(val)
                else:
                    raise EpanetMsxException(403, "at line {} of [OPTIONS] section:\n{}".format(lnum, line))
            except ValueError:
                raise EpanetMsxException(404, "at line {} of [OPTIONS] section:\n{}".format(lnum, line))
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [OPTIONS] section:\n{}".format(lnum, line)) from e

    def _read_species(self):
        note = ENcomment()
        for lnum, line in self.sections["[SPECIES]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) < 3:
                    raise EpanetMsxException(402, "at line {} of [SPECIES] section:\n{}".format(lnum, line))
                try:
                    typ = SpeciesType.get(vals[0], allow_none=False)
                except ValueError as e:
                    raise MSXValueError(403, vals[0], "at line {} of [SPECIES] section:\n{}".format(lnum, line)) from e
                if len(vals) == 3:
                    self.rxn.add_species(vals[1], typ, vals[2], note=note)
                elif len(vals) == 5:
                    self.rxn.add_species(vals[1], typ, vals[2], float(vals[3]), float(vals[4]), note=note)
                else:
                    raise EpanetMsxException(201, "at line {} of [SPECIES] section:\n{}".format(lnum, line))
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [SPECIES] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_coefficients(self):
        note = ENcomment()
        for lnum, line in self.sections["[COEFFICIENTS]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) < 3:
                    raise EpanetMsxException(402, "at line {} of [COEFFICIENTS] section:\n{}".format(lnum, line))
                typ = VariableType.get(vals[0], allow_none=False)
                if typ is VariableType.CONSTANT:
                    self.rxn.add_constant(vals[1], float(vals[2]), note=note)
                elif typ is VariableType.PARAMETER:
                    self.rxn.add_parameter(vals[1], float(vals[2]), note=note)
                else:
                    raise MSXValueError(403, vals[0], "at line {} of [COEFFICIENTS] section:\n{}".format(lnum, line))
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [COEFFICIENTS] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_terms(self):
        note = ENcomment()
        for lnum, line in self.sections["[TERMS]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) < 2:
                    raise SyntaxError("Invalid [TERMS] entry")
                self.rxn.add_term(vals[0], " ".join(vals[1:]), note=note)
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [TERMS] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_pipes(self):
        note = ENcomment()
        for lnum, line in self.sections["[PIPES]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) < 3:
                    raise SyntaxError("Invalid [PIPES] entry")
                reaction = self.rxn.add_reaction(vals[1], "pipe", vals[0], " ".join(vals[2:]), note=note)
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [PIPES] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_tanks(self):
        note = ENcomment()
        for lnum, line in self.sections["[TANKS]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) < 3:
                    raise SyntaxError("Invalid [TANKS] entry")
                self.rxn.add_reaction(vals[1], "tank", vals[0], " ".join(vals[2:]), note=note)
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [TANKS] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_source_dict(self):
        note = ENcomment()
        for lnum, line in self.sections["[SOURCES]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) == 4:
                    typ, node, spec, strength = vals
                    pat = None
                else:
                    typ, node, spec, strength, pat = vals
                if spec not in self.rxn.reaction_system:
                    raise ValueError("Undefined species in [SOURCES] section: {}".format(spec))
                if spec not in self.rxn.network_data.sources.keys():
                    self.rxn.network_data.sources[spec] = dict()
                source = dict(source_type=typ, strength=strength, pattern=pat, note=note)
                self.rxn.network_data.sources[spec][node] = source
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [SOURCES] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_quality(self):
        note = ENcomment()
        for lnum, line in self.sections["[QUALITY]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) == 4:
                    cmd, netid, spec, concen = vals
                else:
                    cmd, spec, concen = vals
                if cmd[0].lower() not in ["g", "n", "l"]:
                    raise SyntaxError("Unknown first word in [QUALITY] section")
                if spec not in self.rxn.reaction_system.species:
                    raise ValueError("Undefined species in [QUALITY] section: {}".format(spec))
                # FIXME: check for existence
                # if spec not in self.rxn.net.initial_quality:
                #     self.rxn.net.new_quality_values(spec)
                #     self.rxn._inital_qual_dict[spec]["global"] = None
                #     self.rxn._inital_qual_dict[spec]["nodes"] = dict()
                #     self.rxn._inital_qual_dict[spec]["links"] = dict()
                if cmd[0].lower() == "g":
                    self.rxn.network_data.initial_quality[spec].global_value = float(concen)
                elif cmd[0].lower() == "n":
                    self.rxn.network_data.initial_quality[spec].node_values[netid] = float(concen)
                elif cmd[1].lower() == "l":
                    self.rxn.network_data.initial_quality[spec].link_values[netid] = float(concen)
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [QUALITY] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_parameters(self):
        note = ENcomment()
        for lnum, line in self.sections["[PARAMETERS]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                typ, netid, paramid, value = vals
                if paramid not in self.rxn.reaction_system.parameters:
                    raise ValueError("Invalid parameter {}".format(paramid))
                value = float(value)
                if typ.lower()[0] == "p":
                    self.rxn.network_data.parameter_values[paramid].pipe_values[netid] = value
                elif typ.lower()[0] == "t":
                    self.rxn.network_data.parameter_values[paramid].tank_values[netid] = value
                else:
                    raise ValueError("Invalid parameter type {}".format(typ))
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [PARAMETERS] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_diffusivity(self):
        note = ENcomment()
        for lnum, line in self.sections["[DIFFUSIVITY]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) != 2:
                    raise SyntaxError("Invalid [DIFFUSIVITIES] entry")
                species = self.rxn.reaction_system.species[vals[0]]
                species.diffusivity = float(vals[1])
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [DIFFUSIVITIES] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _read_patterns(self):
        _patterns = dict()
        for lnum, line in self.sections["[PATTERNS]"]:
            # read the lines for each pattern -- patterns can be multiple lines of arbitrary length
            line = line.split(";")[0]
            current = line.split()
            if current == []:
                continue
            pattern_name = current[0]
            if pattern_name not in _patterns:
                _patterns[pattern_name] = []
                for i in current[1:]:
                    _patterns[pattern_name].append(float(i))
            else:
                for i in current[1:]:
                    _patterns[pattern_name].append(float(i))
        for pattern_name, pattern in _patterns.items():
            # add the patterns to the water newtork model
            self.rxn.network_data.add_pattern(pattern_name, pattern)

    def _read_report(self):
        note = ENcomment()
        for lnum, line in self.sections["[REPORT]"]:
            vals, comment = _split_line(line)
            if vals is None:
                if comment is not None:
                    note.pre.append(comment)
                continue
            try:
                if comment is not None:
                    note.post = comment
                if len(vals) == 0:
                    continue
                if len(vals) < 2:
                    raise SyntaxError("Invalid number of arguments in [REPORT] section")
                cmd = vals[0][0].lower()
                if cmd == "n":  # NODES
                    if self.rxn._options.report.nodes is None:
                        if vals[1].upper() == "ALL":
                            self.rxn._options.report.nodes = "ALL"
                        else:
                            self.rxn._options.report.nodes = list()
                            self.rxn._options.report.nodes.extend(vals[1:])
                    elif isinstance(self.rxn._options.report.nodes, list):
                        if vals[1].upper() == "ALL":
                            self.rxn._options.report.nodes = "ALL"
                        else:
                            self.rxn._options.report.nodes.extend(vals[1:])
                elif cmd == "l":  # LINKS
                    if self.rxn._options.report.links is None:
                        if vals[1].upper() == "ALL":
                            self.rxn._options.report.links = "ALL"
                        else:
                            self.rxn._options.report.links = list()
                            self.rxn._options.report.links.extend(vals[1:])
                    elif isinstance(self.rxn._options.report.links, list):
                        if vals[1].upper() == "ALL":
                            self.rxn._options.report.links = "ALL"
                        else:
                            self.rxn._options.report.links.extend(vals[1:])
                elif cmd == "f":
                    self.rxn._options.report.report_filename = vals[1]
                elif cmd == "p":
                    self.rxn._options.report.pagesize = vals[1]
                elif cmd == "s":
                    if vals[1] not in self.rxn.reaction_system.species:
                        raise ValueError("Undefined species in [REPORT] section: {}".format(vals[1]))
                    self.rxn._options.report.species[vals[1]] = True if vals[2].lower().startswith("y") else False
                    if len(vals) == 4:
                        self.rxn._options.report.species_precision[vals[1]] = int(vals[3])
                else:
                    raise SyntaxError("Invalid syntax in [REPORT] section: unknown first word")
            except EpanetMsxException:
                raise
            except Exception as e:
                raise EpanetMsxException(201, "at line {} of [REPORT] section:\n{}".format(lnum, line)) from e
            else:
                note = ENcomment()

    def _write_title(self, fout):
        fout.write("[TITLE]\n")
        fout.write("  {}\n".format(self.rxn.title))
        fout.write("\n")

    def _write_options(self, fout):
        opts = self.rxn._options
        fout.write("[OPTIONS]\n")
        fout.write("  AREA_UNITS  {}\n".format(opts.area_units.upper()))
        fout.write("  RATE_UNITS  {}\n".format(opts.rate_units.upper()))
        fout.write("  SOLVER      {}\n".format(opts.solver.upper()))
        fout.write("  COUPLING    {}\n".format(opts.coupling.upper()))
        fout.write("  TIMESTEP    {}\n".format(opts.timestep))
        fout.write("  ATOL        {}\n".format(opts.atol))
        fout.write("  RTOL        {}\n".format(opts.rtol))
        fout.write("  COMPILER    {}\n".format(opts.compiler.upper()))
        fout.write("  SEGMENTS    {}\n".format(opts.segments))
        fout.write("  PECLET      {}\n".format(opts.peclet))
        fout.write("\n")

    def _write_species(self, fout):
        fout.write("[SPECIES]\n")

        def to_msx_string(spec: Species) -> str:
            tols = spec.get_tolerances()
            if tols is None:
                tolstr = ""
            else:
                tolstr = " {:12.6g} {:12.6g}".format(*tols)
            return "{:<12s} {:<8s} {:<8s}{:s}".format(
                spec.species_type.name.upper(),
                spec.name,
                spec.units,
                tolstr,
            )

        for var in self.rxn.reaction_system.species.values():
            if isinstance(var.note, ENcomment):
                fout.write("{}\n".format(var.note.wrap_msx_string(to_msx_string(var))))
            elif isinstance(var.note, str):
                fout.write("  {} ; {}\n".format(to_msx_string(var), var.note))
            else:
                fout.write("  {}\n".format(to_msx_string(var)))
        fout.write("\n")

    def _write_coefficients(self, fout):
        fout.write("[COEFFICIENTS]\n")

        def to_msx_string(coeff: Union[Constant, Parameter]) -> str:
            # if self.units is not None:
            #     post = r' ; {"units"="' + str(self.units) + r'"}'
            # else:
            post = ""
            return "{:<12s} {:<8s} {:<16s}{}".format(
                coeff.var_type.name.upper(),
                coeff.name,
                str(coeff.global_value if isinstance(coeff, Parameter) else coeff.value),
                post,
            )

        for var in self.rxn.reaction_system.constants.values():
            if isinstance(var.note, ENcomment):
                fout.write("{}\n".format(var.note.wrap_msx_string(to_msx_string(var))))
            elif isinstance(var.note, str):
                fout.write("  {} ; {}\n".format(to_msx_string(var), var.note))
            else:
                fout.write("  {}\n".format(to_msx_string(var)))
        for var in self.rxn.reaction_system.parameters.values():
            if isinstance(var.note, ENcomment):
                fout.write("{}\n".format(var.note.wrap_msx_string(to_msx_string(var))))
            elif isinstance(var.note, str):
                fout.write("  {} ; {}\n".format(to_msx_string(var), var.note))
            else:
                fout.write("  {}\n".format(to_msx_string(var)))
        fout.write("\n")

    def _write_terms(self, fout):
        fout.write("[TERMS]\n")

        def to_msx_string(term: Term) -> str:
            return "{:<8s} {:<64s}".format(term.name, term.expression)

        for var in self.rxn.reaction_system.terms.values():
            if isinstance(var.note, ENcomment):
                fout.write("{}\n".format(var.note.wrap_msx_string(to_msx_string(var))))
            elif isinstance(var.note, str):
                fout.write("  {} ; {}\n".format(to_msx_string(var), var.note))
            else:
                fout.write("  {}\n".format(to_msx_string(var)))
        fout.write("\n")

    def _write_pipes(self, fout):
        fout.write("[PIPES]\n")
        for spec in self.rxn.reaction_system.species.values():
            var = spec.pipe_reaction
            if var is None:
                raise MSXValueError(507, note=" species {}".format(str(spec)))
            if isinstance(var.note, ENcomment):
                fout.write(
                    "{}\n".format(
                        var.note.wrap_msx_string(
                            "{:<12s} {:<8s} {:<32s}".format(var.expression_type.name.upper(), str(var.species_name), var.expression)
                        )
                    )
                )
            elif isinstance(var.note, str):
                fout.write(
                    "  {} ; {}\n".format(
                        "{:<12s} {:<8s} {:<32s}".format(var.expression_type.name.upper(), str(var.species_name), var.expression),
                        var.note,
                    )
                )
            else:
                fout.write(
                    "  {}\n".format(
                        "{:<12s} {:<8s} {:<32s}".format(var.expression_type.name.upper(), str(var.species_name), var.expression)
                    )
                )
        fout.write("\n")

    def _write_tanks(self, fout):
        fout.write("[TANKS]\n")
        for spec in self.rxn.reaction_system.species.values():
            if spec.species_type.name == 'WALL':
                continue
            try:
                var = spec.tank_reaction
            except KeyError:
                logger.warn('Species {} does not have a tank reaction - this may be a problem'.format(str(spec)))
                continue
            if var is None:
                raise MSXValueError(508, note=" species {}".format(str(spec)))
            if isinstance(var.note, ENcomment):
                fout.write(
                    "{}\n".format(
                        var.note.wrap_msx_string(
                            "{:<12s} {:<8s} {:<32s}".format(var.expression_type.name.upper(), str(var.species_name), var.expression)
                        )
                    )
                )
            elif isinstance(var.note, str):
                fout.write(
                    "  {} ; {}\n".format(
                        "{:<12s} {:<8s} {:<32s}".format(var.expression_type.name.upper(), str(var.species_name), var.expression),
                        var.note,
                    )
                )
            else:
                fout.write(
                    "  {}\n".format(
                        "{:<12s} {:<8s} {:<32s}".format(var.expression_type.name.upper(), str(var.species_name), var.expression)
                    )
                )
        fout.write("\n")

    def _write_source_dict(self, fout):
        fout.write("[SOURCES]\n")
        for species in self.rxn.network_data.sources.keys():
            for node, src in self.rxn.network_data.sources[species].items():
                if isinstance(src["note"], ENcomment):
                    fout.write(
                        src["note"].wrap_msx_string(
                            "{:<10s} {:<8s} {:<8s} {:12s} {:<12s}".format(
                                src["source_type"],
                                node,
                                species,
                                src["strength"],
                                src["pattern"] if src["pattern"] is not None else "",
                            )
                        )
                    )
                elif isinstance(src["note"], str):
                    fout.write(
                        "  {:<10s} {:<8s} {:<8s} {} {:<12s} ; {}\n".format(
                            src["source_type"],
                            node,
                            species,
                            src["strength"],
                            src["pattern"] if src["pattern"] is not None else "",
                            src["note"],
                        )
                    )
                else:
                    fout.write(
                        "  {:<10s} {:<8s} {:<8s} {} {:<12s}\n".format(
                            src["source_type"],
                            node,
                            species,
                            src["strength"],
                            src["pattern"] if src["pattern"] is not None else "",
                        )
                    )
                if src["note"] is not None:
                    fout.write("\n")
        fout.write("\n")

    def _write_quality(self, fout):
        fout.write("[QUALITY]\n")
        for species, val in self.rxn.network_data.initial_quality.items():
            if val.global_value:
                fout.write("  {:<8s} {:<8s} {}\n".format("GLOBAL", species, val.global_value))
            for typ in ["node_values", "link_values"]:
                for node, conc in getattr(val, typ).items():
                    fout.write("  {:<8s} {:<8s} {:<8s} {}\n".format(typ.upper()[0:4], node, species, conc))
        fout.write("\n")

    def _write_parameters(self, fout):
        fout.write("[PARAMETERS]\n")
        for name, var in self.rxn.network_data.parameter_values.items():
            had_entries = False
            for pipeID, value in var.pipe_values.items():
                fout.write("  PIPE     {:<8s} {:<8s} {}\n".format(pipeID, name, value))
                had_entries = True
            for tankID, value in var.tank_values.items():
                fout.write("  PIPE     {:<8s} {:<8s} {}\n".format(tankID, name, value))
                had_entries = True
            if had_entries:
                fout.write("\n")
        fout.write("\n")

    def _write_patterns(self, fout):
        fout.write("[PATTERNS]\n")
        for pattern_name, pattern in self.rxn.network_data.patterns.items():
            num_columns = 10
            count = 0
            for i in pattern:  # .multipliers:
                if count % num_columns == 0:
                    fout.write("\n  {:<8s} {:g}".format(pattern_name, i))
                else:
                    fout.write("  {:g}".format(i))
                count += 1
            fout.write("\n")
        fout.write("\n")

    def _write_diffusivity(self, fout):
        fout.write("[DIFFUSIVITY]\n")
        for name in self.rxn.species_name_list:
            spec: Species = self.rxn.reaction_system.species[name]
            if spec.diffusivity is not None:
                fout.write("  {:<8s} {}\n".format(name, spec.diffusivity))
        fout.write("\n")

    def _write_report(self, fout):
        fout.write("[REPORT]\n")
        if self.rxn._options.report.nodes is not None:
            if isinstance(self.rxn._options.report.nodes, str):
                fout.write("  NODES     {}\n".format(self.rxn.options.report.nodes))
            else:
                fout.write("  NODES     {}\n".format(" ".join(self.rxn.options.report.nodes)))
        if self.rxn._options.report.links is not None:
            if isinstance(self.rxn._options.report.links, str):
                fout.write("  LINKS     {}\n".format(self.rxn.options.report.links))
            else:
                fout.write("  LINKS     {}\n".format(" ".join(self.rxn.options.report.links)))
        for spec, val in self.rxn._options.report.species.items():
            fout.write(
                "  SPECIES   {:<8s} {:<3s} {}\n".format(
                    spec,
                    "YES" if val else "NO",
                    self.rxn._options.report.species_precision[spec]
                    if spec in self.rxn._options.report.species_precision.keys()
                    else "",
                )
            )
        if self.rxn._options.report.report_filename:
            fout.write("  FILE      {}\n".format(self.rxn._options.report.report_filename))
        if self.rxn._options.report.pagesize:
            fout.write("  PAGESIZE  {}\n".format(self.rxn._options.report.pagesize))
        fout.write("\n")


def MsxBinFile(filename, wn, res = None):
    duration = int(wn.options.time.duration)
    if res is None:
        from wntr.sim.results import SimulationResults
        res = SimulationResults()
        res.node = {}
        res.link = {}
    with open(filename, "rb") as fin:
        ftype = "=f4"
        idlen = 32
        prolog = np.fromfile(fin, dtype=np.int32, count=6)
        magic1 = prolog[0]
        version = prolog[1]
        nnodes = prolog[2]
        nlinks = prolog[3]
        nspecies = prolog[4]
        reportstep = prolog[5]
        species_list = []
        node_list = wn.node_name_list
        link_list = wn.link_name_list
        # print(magic1, version, nnodes, nlinks, nspecies, reportstep)

        species_mass = []
        for i in range(nspecies):
            species_len = np.fromfile(fin, dtype=np.int32, count=1)[0]
            # print(species_len)
            species_name = "".join(chr(f) for f in np.fromfile(fin, dtype=np.uint8, count=species_len) if f != 0)
            # print(species_name)
            species_list.append(species_name)
            species_mass.append("".join(chr(f) for f in np.fromfile(fin, dtype=np.uint8, count=16) if f != 0))
            
        timerange = range(0, duration + 1, reportstep)

        tr = len(timerange)

        row1 = ["node"] * nnodes * len(species_list) + ["link"] * nlinks * len(species_list)
        row2 = []
        for i in [nnodes, nlinks]:
            for j in species_list:
                row2.append([j] * i)
        row2 = [y for x in row2 for y in x]
        row3 = [node_list for i in species_list] + [link_list for i in species_list]
        row3 = [y for x in row3 for y in x]

        tuples = list(zip(row1, row2, row3))
        index = pd.MultiIndex.from_tuples(tuples, names=["type", "species", "name"])

        try:
            data = np.fromfile(fin, dtype=np.dtype(ftype), count=tr * (len(species_list * (nnodes + nlinks))))
            data = np.reshape(data, (tr, len(species_list * (nnodes + nlinks))))
        except Exception as e:
            print(e)
            print("oops")

        postlog = np.fromfile(fin, dtype=np.int32, count=4)
        offset = postlog[0]
        numreport = postlog[1]
        errorcode = postlog[2]
        magicnew = postlog[3]
        if errorcode != 0:
            print(f"ERROR CODE: {errorcode}")
            print(offset, numreport)

        df_fin = pd.DataFrame(index=index, columns=timerange).transpose()
        if magic1 == magicnew:
            # print("Magic# Match")
            df_fin = pd.DataFrame(data.transpose(), index=index, columns=timerange)
            df_fin = df_fin.transpose()

        else:
            print("Magic#s do not match!")
    for species in species_list:
        res.node[species] = df_fin['node'][species]
        res.link[species] = df_fin['link'][species]
    return res



"""Model parameters for the WNTRSimulator."""

import logging
from wntr.sim import aml
from wntr.utils.polynomial_interpolation import cubic_spline
import math
from wntr.network import LinkStatus
from wntr.sim.models.utils import ModelUpdater, Definition


logger = logging.getLogger(__name__)


def source_head_param(m, wn):
    """
    Add a head param to the model

    Parameters
    ----------
    m: wntr.sim.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    """
    if not hasattr(m, 'source_head'):
        m.source_head = aml.ParamDict()

        for node_name, node in wn.tanks():
            m.source_head[node_name] = aml.Param(node.head)
        for node_name, node in wn.reservoirs():
            m.source_head[node_name] = aml.Param(node.head_timeseries.at(wn.sim_time))
    else:
        for node_name, node in wn.tanks():
            m.source_head[node_name].value = node.head
        for node_name, node in wn.reservoirs():
            m.source_head[node_name].value = node.head_timeseries.at(wn.sim_time)


def expected_demand_param(m, wn):
    """
    Add a demand parameter to the model

    Parameters
    ----------
    m: wntr.sim.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    """
    demand_multiplier = wn.options.hydraulic.demand_multiplier
    pattern_start = wn.options.time.pattern_start
    
    if not hasattr(m, 'expected_demand'):
        m.expected_demand = aml.ParamDict()

        for node_name, node in wn.junctions():
            m.expected_demand[node_name] = aml.Param(node.demand_timeseries_list.at(wn.sim_time+pattern_start, multiplier=demand_multiplier))
    else:
        for node_name, node in wn.junctions():
            m.expected_demand[node_name].value = node.demand_timeseries_list.at(wn.sim_time+pattern_start, multiplier=demand_multiplier)


class pmin_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a minimum pressure parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of junction names
        """
        if not hasattr(m, 'pmin'):
            m.pmin = aml.ParamDict()

        if index_over is None:
            index_over = wn.junction_name_list

        for node_name in index_over:
            node = wn.get_node(node_name)
            if node.minimum_pressure is None:
                minimum_pressure = wn.options.hydraulic.minimum_pressure
            else:
                minimum_pressure = node.minimum_pressure
                
            if node_name in m.pmin:
                m.pmin[node_name].value = minimum_pressure
            else:
                m.pmin[node_name] = aml.Param(minimum_pressure)

            updater.add(node, 'minimum_pressure', pmin_param.update)


class pnom_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a required pressure parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of junction names
        """
        if not hasattr(m, 'pnom'):
            m.pnom = aml.ParamDict()

        if index_over is None:
            index_over = wn.junction_name_list

        for node_name in index_over:
            node = wn.get_node(node_name)
            if node.required_pressure is None:
                required_pressure = wn.options.hydraulic.required_pressure
            else:
                required_pressure = node.required_pressure
                
            if required_pressure <= m.pdd_smoothing_delta:
                raise ValueError('Required pressure for node %s must be greater than %s, the smoothing delta', node_name, m.pdd_smoothing_delta)
            if node_name in m.pnom:
                m.pnom[node_name].value = required_pressure
            else:
                m.pnom[node_name] = aml.Param(required_pressure)

            updater.add(node, 'required_pressure', pnom_param.update)


class leak_coeff_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a leak discharge coefficient parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of junction/tank names
        """
        if not hasattr(m, 'leak_coeff'):
            m.leak_coeff = aml.ParamDict()

        if index_over is None:
            index_over = wn.junction_name_list + wn.tank_name_list

        for node_name in index_over:
            node = wn.get_node(node_name)
            if node_name in m.leak_coeff:
                m.leak_coeff[node_name].value = node.leak_discharge_coeff
            else:
                m.leak_coeff[node_name] = aml.Param(node.leak_discharge_coeff)

            updater.add(node, 'leak_discharge_coeff', leak_coeff_param.update)


class leak_area_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a leak discharge coefficient parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of junction/tank names
        """
        if not hasattr(m, 'leak_area'):
            m.leak_area = aml.ParamDict()

        if index_over is None:
            index_over = wn.junction_name_list + wn.tank_name_list

        for node_name in index_over:
            node = wn.get_node(node_name)
            if node_name in m.leak_area:
                m.leak_area[node_name].value = node.leak_area
            else:
                m.leak_area[node_name] = aml.Param(node.leak_area)

            updater.add(node, 'leak_area', leak_area_param.update)


class pdd_poly_coeffs_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add parameters to the model for pdd smoothing polynomial coefficients

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of junction names
        """
        if not hasattr(m, 'pdd_poly1_coeffs_a'):
            m.pdd_poly1_coeffs_a = aml.ParamDict()
            m.pdd_poly1_coeffs_b = aml.ParamDict()
            m.pdd_poly1_coeffs_c = aml.ParamDict()
            m.pdd_poly1_coeffs_d = aml.ParamDict()
            m.pdd_poly2_coeffs_a = aml.ParamDict()
            m.pdd_poly2_coeffs_b = aml.ParamDict()
            m.pdd_poly2_coeffs_c = aml.ParamDict()
            m.pdd_poly2_coeffs_d = aml.ParamDict()

        if index_over is None:
            index_over = wn.junction_name_list

        for node_name in index_over:
            node = wn.get_node(node_name)
            if node.minimum_pressure is None:
                pmin = wn.options.hydraulic.minimum_pressure
            else:
                pmin = node.minimum_pressure
            if node.required_pressure is None:
                pnom = wn.options.hydraulic.required_pressure
            else:
                pnom = node.required_pressure
            x1 = pmin
            f1 = 0.0
            x2 = pmin + m.pdd_smoothing_delta
            f2 = ((x2 - pmin)/(pnom-pmin))**0.5
            df1 = m.pdd_slope
            df2 = 0.5*((x2-pmin)/(pnom-pmin))**(-0.5)*1.0/(pnom-pmin)
            a1, b1, c1, d1 = cubic_spline(x1, x2, f1, f2, df1, df2)
            x1 = pnom-m.pdd_smoothing_delta
            f1 = ((x1-pmin)/(pnom-pmin))**0.5
            x2 = pnom
            f2 = 1.0
            df1 = 0.5*((x1-pmin)/(pnom-pmin))**(-0.5)*1.0/(pnom-pmin)
            df2 = m.pdd_slope
            a2, b2, c2, d2 = cubic_spline(x1, x2, f1, f2, df1, df2)
            if node_name in m.pdd_poly1_coeffs_a:
                m.pdd_poly1_coeffs_a[node_name].value = a1
                m.pdd_poly1_coeffs_b[node_name].value = b1
                m.pdd_poly1_coeffs_c[node_name].value = c1
                m.pdd_poly1_coeffs_d[node_name].value = d1
                m.pdd_poly2_coeffs_a[node_name].value = a2
                m.pdd_poly2_coeffs_b[node_name].value = b2
                m.pdd_poly2_coeffs_c[node_name].value = c2
                m.pdd_poly2_coeffs_d[node_name].value = d2
            else:
                m.pdd_poly1_coeffs_a[node_name] = aml.Param(a1)
                m.pdd_poly1_coeffs_b[node_name] = aml.Param(b1)
                m.pdd_poly1_coeffs_c[node_name] = aml.Param(c1)
                m.pdd_poly1_coeffs_d[node_name] = aml.Param(d1)
                m.pdd_poly2_coeffs_a[node_name] = aml.Param(a2)
                m.pdd_poly2_coeffs_b[node_name] = aml.Param(b2)
                m.pdd_poly2_coeffs_c[node_name] = aml.Param(c2)
                m.pdd_poly2_coeffs_d[node_name] = aml.Param(d2)

            updater.add(node, 'minimum_pressure', pdd_poly_coeffs_param.update)
            updater.add(node, 'required_pressure', pdd_poly_coeffs_param.update)


class leak_poly_coeffs_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add parameters to the model for leak smoothing polynomial coefficients

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of junction names
        """
        if not hasattr(m, 'leak_poly_coeffs_a'):
            m.leak_poly_coeffs_a = aml.ParamDict()
            m.leak_poly_coeffs_b = aml.ParamDict()
            m.leak_poly_coeffs_c = aml.ParamDict()
            m.leak_poly_coeffs_d = aml.ParamDict()

        if index_over is None:
            index_over = wn.junction_name_list + wn.tank_name_list

        for node_name in index_over:
            node = wn.get_node(node_name)
            x1 = 0.0
            f1 = 0.0
            x2 = x1 + m.leak_delta
            f2 = node.leak_discharge_coeff*node.leak_area*(2.0*9.81*x2)**0.5
            df1 = m.leak_slope
            df2 = 0.5*node.leak_discharge_coeff*node.leak_area*(2.0*9.81)**0.5*x2**(-0.5)
            a, b, c, d = cubic_spline(x1, x2, f1, f2, df1, df2)
            if node_name in m.leak_poly_coeffs_a:
                m.leak_poly_coeffs_a[node_name].value = a
                m.leak_poly_coeffs_b[node_name].value = b
                m.leak_poly_coeffs_c[node_name].value = c
                m.leak_poly_coeffs_d[node_name].value = d
            else:
                m.leak_poly_coeffs_a[node_name] = aml.Param(a)
                m.leak_poly_coeffs_b[node_name] = aml.Param(b)
                m.leak_poly_coeffs_c[node_name] = aml.Param(c)
                m.leak_poly_coeffs_d[node_name] = aml.Param(d)

            updater.add(node, 'leak_discharge_coeff', leak_poly_coeffs_param.update)
            updater.add(node, 'leak_area', leak_poly_coeffs_param.update)


class elevation_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add an elevation parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of junction names
        """
        if not hasattr(m, 'elevation'):
            m.elevation = aml.ParamDict()

        if index_over is None:
            index_over = wn.junction_name_list

        for node_name in index_over:
            node = wn.get_node(node_name)
            if node_name in m.elevation:
                m.elevation[node_name].value = node.elevation
            else:
                m.elevation[node_name] = aml.Param(node.elevation)

            updater.add(node, 'elevation', elevation_param.update)


class hw_resistance_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a HW resistance coefficient parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of pipe names
        """
        if not hasattr(m, 'hw_resistance'):
            m.hw_resistance = aml.ParamDict()

        if index_over is None:
            index_over = wn.pipe_name_list

        for link_name in index_over:
            link = wn.get_link(link_name)
            value = m.hw_k * link.roughness**(-1.852) * link.diameter**(-4.871) * link.length
            if link_name in m.hw_resistance:
                m.hw_resistance[link_name].value = value
            else:
                m.hw_resistance[link_name] = aml.Param(value)

            updater.add(link, 'roughness', hw_resistance_param.update)
            updater.add(link, 'diameter', hw_resistance_param.update)
            updater.add(link, 'length', hw_resistance_param.update)


class minor_loss_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a minor loss coefficient parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of pipe names
        """
        if not hasattr(m, 'minor_loss'):
            m.minor_loss = aml.ParamDict()

        if index_over is None:
            index_over = wn.pipe_name_list + wn.valve_name_list

        for link_name in index_over:
            link = wn.get_link(link_name)
            value = 8.0 * link.minor_loss / (9.81 * math.pi**2 * link.diameter**4)
            if link_name in m.minor_loss:
                m.minor_loss[link_name].value = value
            else:
                m.minor_loss[link_name] = aml.Param(value)

            updater.add(link, 'minor_loss', minor_loss_param.update)
            updater.add(link, 'diameter', minor_loss_param.update)


class tcv_resistance_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a tcv resistance parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of tcv names
        """
        if not hasattr(m, 'tcv_resistance'):
            m.tcv_resistance = aml.ParamDict()

        if index_over is None:
            index_over = wn.tcv_name_list

        for link_name in index_over:
            link = wn.get_link(link_name)
            value = 8.0 * link.setting / (9.81 * math.pi**2 * link.diameter**4)
            if link_name in m.tcv_resistance:
                m.tcv_resistance[link_name].value = value
            else:
                m.tcv_resistance[link_name] = aml.Param(value)

            updater.add(link, 'setting', tcv_resistance_param.update)
            updater.add(link, 'diameter', tcv_resistance_param.update)


class pump_power_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a power parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of pump names
        """
        if not hasattr(m, 'pump_power'):
            m.pump_power = aml.ParamDict()

        if index_over is None:
            index_over = wn.power_pump_name_list

        for link_name in index_over:
            link = wn.get_link(link_name)
            value = link.power
            if link_name in m.pump_power:
                m.pump_power[link_name].value = value
            else:
                m.pump_power[link_name] = aml.Param(value)

            updater.add(link, 'power', pump_power_param.update)


class valve_setting_param(Definition):
    @classmethod
    def build(cls, m, wn, updater, index_over=None):
        """
        Add a valve setting parameter to the model

        Parameters
        ----------
        m: wntr.sim.aml.aml.Model
        wn: wntr.network.model.WaterNetworkModel
        updater: ModelUpdater
        index_over: list of str
            list of valve names
        """
        if not hasattr(m, 'valve_setting'):
            m.valve_setting = aml.ParamDict()

        if index_over is None:
            index_over = wn.valve_name_list

        for link_name in index_over:
            link = wn.get_link(link_name)
            value = link.setting
            if link_name in m.valve_setting:
                m.valve_setting[link_name].value = value
            else:
                m.valve_setting[link_name] = aml.Param(value)

            updater.add(link, 'setting', valve_setting_param.update)



# coding: utf-8
"""
The wntr.msx package contains methods to define EPANET Multi-species Extension 
(MSX) water quality models.
"""

from .base import VariableType, SpeciesType, ReactionType, ExpressionType
from .elements import Species, Constant, Parameter, Term, Reaction, HydraulicVariable, MathFunction
from .model import MsxModel
from .options import MsxSolverOptions

from . import base, elements, model, options, io



import unittest
from os.path import abspath, dirname, join
from pandas.testing import assert_frame_equal, assert_series_equal

import wntr

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")

class TestTodiniMetrics(unittest.TestCase):
    
    @classmethod
    def setUpClass(self):
        import wntr

        # Network models from Todini, 2000 (based on Abebe and Solomatine, 
        # 1988 - defined h_star)
        self.h_star = 30 # m
        
    def test_Todini_Fig2_optCost_GPM(self):
        inp_file = join(datadir, "Todini_Fig2_optCost_GPM.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        sim = wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        # Compute todini index
        head = results.node["head"]
        pressure = results.node["pressure"]
        demand = results.node["demand"]
        flowrate = results.link["flowrate"]
        todini = wntr.metrics.todini_index(
            head, pressure, demand, flowrate, wn, self.h_star
        )  

        expected = 0.22
        error = abs(todini[0] - expected)
        # print(todini[0], expected, error)
        self.assertLess(error, 0.01)

    def test_Todini_Fig2_optCost_CMH(self):
        inp_file = join(datadir, "Todini_Fig2_optCost_CMH.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        sim = wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        # Compute todini index
        head = results.node["head"]
        pressure = results.node["pressure"]
        demand = results.node["demand"]
        flowrate = results.link["flowrate"]
        todini = wntr.metrics.todini_index(
            head, pressure, demand, flowrate, wn, self.h_star
        )  

        expected = 0.22
        error = abs(todini[0] - expected)
        # print(todini[0], expected, error)
        self.assertLess(error, 0.01)

    def test_Todini_Fig2_solA_GPM(self):
        inp_file = join(datadir, "Todini_Fig2_solA_GPM.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        sim = wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        # Compute todini index
        head = results.node["head"]
        pressure = results.node["pressure"]
        demand = results.node["demand"]
        flowrate = results.link["flowrate"]
        todini = wntr.metrics.todini_index(
            head, pressure, demand, flowrate, wn, self.h_star
        )  

        expected = 0.41
        error = abs(todini[0] - expected)
        # print(todini[0], expected, error)
        self.assertLess(error, 0.03)

    def test_Todini_Fig2_solA_CMH(self):
        inp_file = join(datadir, "Todini_Fig2_solA_CMH.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        sim = wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        # Compute todini index
        head = results.node["head"]
        pressure = results.node["pressure"]
        demand = results.node["demand"]
        flowrate = results.link["flowrate"]
        todini = wntr.metrics.todini_index(
            head, pressure, demand, flowrate, wn, self.h_star
        ) 

        expected = 0.41
        error = abs(todini[0] - expected)
        # print(todini[0], expected, error)
        self.assertLess(error, 0.03)

class TestMRIMetric(unittest.TestCase):
    
    @classmethod
    def setUpClass(self):
        inp_file = join(ex_datadir, "Net3.inp")
        self.wn = wntr.network.WaterNetworkModel(inp_file)
        self.wn.options.time.duration = 2*24*3600 # 2 days
        sim = wntr.sim.WNTRSimulator(self.wn)
        self.results = sim.run_sim()
    
    @classmethod
    def tearDownClass(self):
        pass
    
    def test_MRI(self):
        
        elevation = self.wn.query_node_attribute('elevation').loc[self.wn.junction_name_list]
        pressure = self.results.node["pressure"].loc[:,self.wn.junction_name_list]
        demand = self.results.node["demand"].loc[:,self.wn.junction_name_list]
        
        temp = demand.sum()
        nzd_nodes = temp[temp>0].index
        
        average_system_pressure = pressure.loc[:,nzd_nodes].mean().mean()
        Pstar = pressure.loc[:,nzd_nodes].min().min()
        #print(Pstar)
        
        mri_per_junction = wntr.metrics.modified_resilience_index(pressure, elevation, Pstar)
        mri_per_junction_nzd = wntr.metrics.modified_resilience_index(pressure.loc[:,nzd_nodes], elevation[nzd_nodes], Pstar)
        
        mri = wntr.metrics.modified_resilience_index(pressure, elevation, Pstar, demand, False)
        mri_nzd = wntr.metrics.modified_resilience_index(pressure.loc[:,nzd_nodes], elevation[nzd_nodes], Pstar, demand.loc[:,nzd_nodes], False)

        # import matplotlib.pylab as plt
        
        # fig, axes = plt.subplots(2,1, figsize=(5,10))
        # pressure.plot(title='Pressure at junctions', legend=False, ax=axes[0])
        # pressure.loc[:,nzd_nodes].plot(title='Pressure at NZD junctions', legend=False, ax=axes[1])
        
        # fig, axes = plt.subplots(2,1, figsize=(5,10))
        # mri_per_junction.plot(title='MRI at junctions', legend=False, ax=axes[0])
        # mri_per_junction_nzd.plot(title='MRI at NZD junctions', legend=False, ax=axes[1])
        
        # fig, axes = plt.subplots(2,1, figsize=(5,10))
        # mri.plot(title='MRI system average', legend=False, ax=axes[0])
        # mri_nzd.plot(title='MRI system average at NZD junctions', legend=False, ax=axes[1])
        
        assert_series_equal(mri, mri_nzd, check_dtype=False) # system mri over all junctions = system average over demand junctions
        assert_frame_equal(mri_per_junction.loc[:,nzd_nodes], mri_per_junction_nzd, check_dtype=False) 
        
        self.assertEqual(mri_per_junction_nzd.min().min(), 0) # the min is 0 because Pstar is the min required pressure
        self.assertLess(mri_per_junction_nzd.max().max(), 1) # for this example, mri per junction is < 1
        
if __name__ == "__main__":
    unittest.main()



"""Generic mathematical solver classes.
"""
import numpy as np
import scipy.sparse as sp
import warnings
import logging
import enum
import time

warnings.filterwarnings(
    "error", "Matrix is exactly singular", sp.linalg.MatrixRankWarning
)
np.set_printoptions(precision=3, threshold=10000, linewidth=300)

logger = logging.getLogger(__name__)


class SolverStatus(enum.IntEnum):
    converged = 1
    error = 0


class NewtonSolver(object):
    """
    Newton Solver class.

    Attributes
    ----------
    log_progress: bool
        If True, the infinity norm of the constraint violation will be logged each iteration
    log_level: int
        The level for logging the infinity norm of the constraint violation
    time_limit: float
        If the wallclock time exceeds time_limit, the newton solver will exit with an error status
    maxiter: int
        If the number of iterations exceeds maxiter, the newton solver will exit with an error status
    tol: float
        The convergence tolerance. If the infinity norm of the constraint violation drops below tol,
        the newton solver will exit with a converged status.
    rho: float
        During the line search, rho is used to reduce the stepsize. It should be strictly between 0 and 1.
    bt_maxiter: int
        The maximum number of line search iterations for each outer iteration
    bt: bool
        If False, a line search will not be used.
    bt_start_iter: int
        A line search will not be used for any iteration prior to bt_start_iter
    """

    def __init__(self, options=None):
        """
        Parameters
        ----------
        options: dict
            A dictionary specifying options for the newton solver. Keys
            should be strings in all caps. See the documentation of the
            NewtonSolver attributes for details on each option. Possible
            keys are:
                | "LOG_PROGRESS" (NewtonSolver.log_progress)
                | "LOG_LEVEL" (NewtonSolver.log_level)
                | "TIME_LIMIT" (NewtonSolver.time_limit)
                | "MAXITER" (NewtonSolver.maxiter)
                | "TOL" (NewtonSolver.tol)
                | "BT_RHO" (NewtonSolver.rho)
                | "BT_MAXITER" (NewtonSolver.bt_maxiter)
                | "BACKTRACKING" (NewtonSolver.bt)
                | "BT_START_ITER" (NewtonSolver.bt_start_iter)
        """
        if options is None:
            options = {}
        self._options = options

        if "LOG_PROGRESS" not in self._options:
            self.log_progress = False
        else:
            self.log_progress = self._options["LOG_PROGRESS"]

        if "LOG_LEVEL" not in self._options:
            self.log_level = logging.DEBUG
        else:
            self.log_level = self._options["LOG_LEVEL"]

        if "TIME_LIMIT" not in self._options:
            self.time_limit = 3600
        else:
            self.time_limit = self._options["TIME_LIMIT"]

        if "MAXITER" not in self._options:
            self.maxiter = 3000
        else:
            self.maxiter = self._options["MAXITER"]

        if "TOL" not in self._options:
            self.tol = 1e-6
        else:
            self.tol = self._options["TOL"]

        if "BT_RHO" not in self._options:
            self.rho = 0.5
        else:
            self.rho = self._options["BT_RHO"]

        if "BT_MAXITER" not in self._options:
            self.bt_maxiter = 100
        else:
            self.bt_maxiter = self._options["BT_MAXITER"]

        if "BACKTRACKING" not in self._options:
            self.bt = True
        else:
            self.bt = self._options["BACKTRACKING"]

        if "BT_START_ITER" not in self._options:
            self.bt_start_iter = 0
        else:
            self.bt_start_iter = self._options["BT_START_ITER"]

    def solve(self, model, ostream=None):
        """

        Parameters
        ----------
        model: wntr.aml.Model

        Returns
        -------
        status: SolverStatus
        message: str
        iter_count: int
        """
        t0 = time.time()

        x = model.get_x()
        if len(x) == 0:
            return (
                SolverStatus.converged,
                "No variables or constraints",
                0,
            )

        use_r_ = False

        # MAIN NEWTON LOOP
        for outer_iter in range(self.maxiter):
            if time.time() - t0 >= self.time_limit:
                return (
                    SolverStatus.error,
                    "Time limit exceeded",
                    outer_iter,
                )

            if use_r_:
                r = r_
                r_norm = new_norm
            else:
                r = model.evaluate_residuals()
                r_norm = np.max(abs(r))

            if self.log_progress or ostream is not None:
                if outer_iter < self.bt_start_iter:
                    msg = f"iter: {outer_iter:<4d} norm: {r_norm:<10.2e} time: {time.time() - t0:<8.4f}"
                    if self.log_progress:
                        logger.log(self.log_level, msg)
                    if ostream is not None:
                        ostream.write(msg + "\n")

            if r_norm < self.tol:
                return (
                    SolverStatus.converged,
                    "Solved Successfully",
                    outer_iter,
                )

            J = model.evaluate_jacobian(x=None)

            # Call Linear solver
            try:
                d = -sp.linalg.spsolve(J, r, permc_spec="COLAMD", use_umfpack=False)
            except sp.linalg.MatrixRankWarning:
                return (
                    SolverStatus.error,
                    "Jacobian is singular at iteration " + str(outer_iter),
                    outer_iter,
                )

            # Backtracking
            alpha = 1.0
            if self.bt and outer_iter >= self.bt_start_iter:
                use_r_ = True
                for iter_bt in range(self.bt_maxiter):
                    x_ = x + alpha * d
                    model.load_var_values_from_x(x_)
                    r_ = model.evaluate_residuals()
                    new_norm = np.max(abs(r_))
                    if new_norm < (1.0 - 0.0001 * alpha) * r_norm:
                        x = x_
                        break
                    else:
                        alpha = alpha * self.rho

                if iter_bt + 1 >= self.bt_maxiter:
                    return (
                        SolverStatus.error,
                        "Line search failed at iteration " + str(outer_iter),
                        outer_iter,
                    )
                if self.log_progress or ostream is not None:
                    msg = f"iter: {outer_iter:<4d} norm: {new_norm:<10.2e} alpha: {alpha:<10.2e} time: {time.time() - t0:<8.4f}"
                    if self.log_progress:
                        logger.log(self.log_level, msg)
                    if ostream is not None:
                        ostream.write(msg + "\n")
            else:
                x += d
                model.load_var_values_from_x(x)

        return (
            SolverStatus.error,
            "Reached maximum number of iterations: " + str(outer_iter),
            outer_iter,
        )



import unittest
from os.path import abspath, dirname, join, exists
import sys, platform

import wntr.epanet.toolkit

if 'darwin' in sys.platform.lower() and 'arm' in platform.platform().lower():
    skip_v2_tests_on_arm = True
else:
    skip_v2_tests_on_arm = False

testdir = dirname(abspath(__file__))
datadir = join(testdir, "..", "..", "examples", "networks")


class TestEpanetToolkit(unittest.TestCase):
    
    def test_isOpen(self):
        for version in [2.0, 2.2,]:
            if version == 2.0 and skip_v2_tests_on_arm:
                continue  # skip v2.0 tests on mac silicon processor
            enData = wntr.epanet.toolkit.ENepanet(version=version)
            enData.inpfile = join(datadir, "Net1.inp")
            self.assertEqual(0, enData.isOpen())
            enData.ENopen(enData.inpfile, "temp.rpt")
            self.assertEqual(1, enData.isOpen())

    def test_ENgetcount(self):
        for version in [2.0, 2.2,]:
            if version == 2.0 and skip_v2_tests_on_arm:
                continue  # skip v2.0 tests on mac silicon processor
            enData = wntr.epanet.toolkit.ENepanet(version=version)
            enData.inpfile = join(datadir, "Net1.inp")
            enData.ENopen(enData.inpfile, "temp.rpt")
            
            nNodes = enData.ENgetcount(wntr.epanet.util.EN.NODECOUNT)
            self.assertEqual(11, nNodes)
            nLinks = enData.ENgetcount(wntr.epanet.util.EN.LINKCOUNT)
            self.assertEqual(13, nLinks)
    
    def test_ENgetflowunits(self):
        for version in [2.0, 2.2,]:
            if version == 2.0 and skip_v2_tests_on_arm:
                continue  # skip v2.0 tests on mac silicon processor
            enData = wntr.epanet.toolkit.ENepanet(version=version)
            enData.inpfile = join(datadir, "Net1.inp")
            enData.ENopen(enData.inpfile, "temp.rpt")
            
            flowunit = enData.ENgetflowunits()
            assert(flowunit==1) # GPM
        
    def test_EN_timeparam(self):
        for version in [2.0, 2.2,]:
            if version == 2.0 and skip_v2_tests_on_arm:
                continue  # skip v2.0 tests on mac silicon processor
            enData = wntr.epanet.toolkit.ENepanet(version=version)
            enData.inpfile = join(datadir, "Net1.inp")
            enData.ENopen(enData.inpfile, "temp.rpt")
            
            duration = enData.ENgettimeparam(0)
            assert(duration==86400) # s
            enData.ENsettimeparam(0, 200)
            duration = enData.ENgettimeparam(0)
            assert(duration==200)
        
    def test_ENgetindex_ENgetvalue(self):
        for version in [2.0, 2.2,]:
            if version == 2.0 and skip_v2_tests_on_arm:
                continue  # skip v2.0 tests on mac silicon processor
            enData = wntr.epanet.toolkit.ENepanet(version=version)
            enData.inpfile = join(datadir, "Net1.inp")
            enData.ENopen(enData.inpfile, "temp.rpt")
            
            node_index = enData.ENgetnodeindex('10') 
            assert(node_index == 1) 
            node_val = enData.ENgetnodevalue(node_index, 0) # ELEVATION = 0
            assert(node_val == 710) 
            enData.ENsetnodevalue(node_index, 0, 170.5)
            node_val = enData.ENgetnodevalue(node_index, 0) # ELEVATION = 0
            assert(node_val == 170.5) 
            
            
            link_index = enData.ENgetlinkindex('11') 
            assert(link_index == 2) 
            link_val = enData.ENgetlinkvalue(link_index, 0) # DIAMETER = 0
            assert(link_val == 14) 
            enData.ENsetlinkvalue(link_index, 0, 16.5)
            link_val = enData.ENgetlinkvalue(link_index, 0) # DIAMETER = 0
            assert(link_val == 16.5) 
        
    def test_ENsaveinpfile(self):
        for version in [2.0, 2.2,]:
            if version == 2.0 and skip_v2_tests_on_arm:
                continue  # skip v2.0 tests on mac silicon processor
            enData = wntr.epanet.toolkit.ENepanet(version=version)
            enData.inpfile = join(datadir, "Net1.inp")
            enData.ENopen(enData.inpfile, "temp.rpt")
            
            enData.ENsaveinpfile("temp_Net1_toolkit.inp") 
            file_exists = exists("temp_Net1_toolkit.inp")
            assert file_exists
        
    def test_runepanet(self):
        inpfile = join(datadir, "Net1.inp")
        wntr.epanet.toolkit.runepanet(inpfile, "temp_runepanet.rpt", "temp_runepanet.bin")
                 
        reader = wntr.epanet.io.BinFile()
        results = reader.read("temp_runepanet.bin")
        
        assert(isinstance(results, wntr.sim.results.SimulationResults))
        assert(results.node['pressure'].shape == (25,11))
        
    def test_runepanet_step(self):
        for version in [2.0, 2.2,]:
            if version == 2.0 and skip_v2_tests_on_arm:
                continue  # skip v2.0 tests on mac silicon processor
            enData = wntr.epanet.toolkit.ENepanet(version=version)
            enData.inpfile = join(datadir, "Net1.inp")
            enData.ENopen(enData.inpfile, "temp_runepanet_step.rpt", "temp_runepanet_step.bin")
            
            enData.ENopenH()
            enData.ENinitH(0)
            t = 0
            while True:
                enData.ENrunH()
                tstep = enData.ENnextH()
                t = t + tstep
                if (tstep <= 0):
                    break
            
            enData.ENcloseH()

            assert (t == 86400)

if __name__ == "__main__":
    unittest.main()



# These tests test controls
import unittest
import warnings
from os.path import abspath, dirname, join
import copy

import wntr
from wntr.epanet.io import _read_control_line

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestValveSettingControls(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        wn = wntr.network.WaterNetworkModel()
        wn.add_reservoir("r1", base_head=10)
        wn.add_junction("j1", base_demand=0)
        wn.add_junction("j2", base_demand=0.05)
        wn.add_pipe("p1", "r1", "j1")
        wn.add_valve("v1", "j1", "j2", valve_type="PRV", 
                     initial_setting=2, initial_status='Active')
        wn.options.time.duration = 3600 * 5
        self.wn = wn
    
    @classmethod
    def tearDownClass(self):
        pass
    
    def test_status_open_when_setting_changes(self):
        wn = copy.deepcopy(self.wn)
        action = wntr.network.ControlAction(
            wn.get_link("v1"), "status", wntr.network.LinkStatus.Closed
        )
        condition = wntr.network.SimTimeCondition(wn, "==", 0)
        control = wntr.network.Control(condition=condition, then_action=action)
        wn.add_control("close_valve", control)

        action = wntr.network.ControlAction(wn.get_link("v1"), "setting", 2)
        condition = wntr.network.SimTimeCondition(wn, "==", 7200)
        control = wntr.network.Control(condition=condition, then_action=action)
        wn.add_control("valve_setting", control)

        sim = wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()
        self.assertEqual(
            results.link["status"].at[7200, "v1"], wntr.network.LinkStatus.Active
        )
    
    def test_initial_status(self):
        # Run simulations with open valve
        wn = copy.deepcopy(self.wn)
        valve_name = 'v1'
        valve = wn.get_link(valve_name)
        valve.initial_status = 'Open'
        
        sim = wntr.sim.EpanetSimulator(wn)
        results_epanet_open = sim.run_sim()
        
        sim = wntr.sim.WNTRSimulator(wn)
        results_wntr_open = sim.run_sim()
        
        # Check that valve is open (1) and flow is not 0
        assert (results_epanet_open.link['status'].loc[:,'v1'] == 1).all()
        assert (results_wntr_open.link['status'].loc[:,'v1'] == 1).all()
        assert (results_epanet_open.link['flowrate'].loc[:,'v1'].abs() > 0).all()
        assert (results_wntr_open.link['flowrate'].loc[:,'v1'].abs() > 0).all()
        
        # Run simulations with closed valve
        wn = copy.deepcopy(self.wn)
        valve_name = 'v1'
        valve = wn.get_link(valve_name)
        valve.initial_status = 'Closed'
        
        sim = wntr.sim.EpanetSimulator(wn)
        results_epanet_closed = sim.run_sim()
        
        sim = wntr.sim.WNTRSimulator(wn)
        results_wntr_closed = sim.run_sim()
        
        # Check that valve is closed (0) and flow is 0
        assert (results_epanet_closed.link['status'].loc[:,'v1'] == 0).all()
        assert (results_wntr_closed.link['status'].loc[:,'v1'] == 0).all()
        assert (results_epanet_closed.link['flowrate'].loc[:,'v1'] == 0).all()
        assert (results_wntr_closed.link['flowrate'].loc[:,'v1'] == 0).all()

    def test_initial_setting(self):
        # Run simulations with valve setting of 4
        wn = copy.deepcopy(self.wn)
        valve_name = 'v1'
        valve = wn.get_link(valve_name)
        # pressure setting on its downstream side when the upstream pressure is above the setting
        valve.initial_setting = 4
        
        sim = wntr.sim.EpanetSimulator(wn)
        results_epanet_open = sim.run_sim()
        
        sim = wntr.sim.WNTRSimulator(wn)
        results_wntr_open = sim.run_sim()
        
        # Check that valve is active (2) and the downstream pressure is 4
        assert (results_epanet_open.link['status'].loc[:,'v1'] == 2).all()
        assert (results_wntr_open.link['status'].loc[:,'v1'] == 2).all()
        assert (results_epanet_open.node['pressure'].loc[:,'j2'] == 4).all()
        assert (results_wntr_open.node['pressure'].loc[:,'j2'] == 4).all()
        
class TestPumpSettingControls(unittest.TestCase):
    def test_status_open_when_setting_changes(self):
        wn = wntr.network.WaterNetworkModel()
        wn.add_reservoir("r1", base_head=10)
        wn.add_junction("j1", base_demand=0)
        wn.add_junction("j2", base_demand=0.1)
        wn.add_junction("j3", base_demand=0.05)
        wn.add_pipe("p1", "r1", "j1")
        wn.add_pipe("p2", "j1", "j3")
        wn.add_curve("pump_curve", "HEAD", [(0.1, 3)])
        wn.add_pump("pump1", "j1", "j2", pump_type="HEAD", pump_parameter="pump_curve")
        wn.options.time.duration = 3600 * 5

        action = wntr.network.ControlAction(
            wn.get_link("pump1"), "status", wntr.network.LinkStatus.Closed
        )
        condition = wntr.network.SimTimeCondition(wn, "==", 0)
        control = wntr.network.Control(condition=condition, then_action=action)
        wn.add_control("close_pump", control)

        action = wntr.network.ControlAction(wn.get_link("pump1"), "base_speed", 1)
        condition = wntr.network.SimTimeCondition(wn, "==", 3600 * 2)
        control = wntr.network.Control(condition=condition, then_action=action)
        wn.add_control("pump_base_speed", control)

        sim = wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()
        self.assertEqual(
            results.link["status"].at[7200, "pump1"], wntr.network.LinkStatus.Open
        )


class TestTimeControls(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

        inp_file = join(test_datadir, "time_controls.inp")
        self.wn = self.wntr.network.WaterNetworkModel(inp_file)
        self.wn.options.time.report_timestep = "all"
        for jname, j in self.wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0

        self.wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(self.wn)
        self.results = sim.run_sim()

    @classmethod
    def tearDownClass(self):
        pass

    def test_time_control_open_vs_closed(self):
        res = self.results
        link_res = res.link
        for t in res.time:
            if t < 5 * 3600 or t >= 10 * 3600:
                self.assertAlmostEqual(
                    link_res["flowrate"].at[t, "pipe2"], 150 / 3600.0
                )
                self.assertEqual(
                    link_res["status"].at[t, "pipe2"], self.wntr.network.LinkStatus.Open
                )
            else:
                self.assertAlmostEqual(link_res["flowrate"].at[t, "pipe2"], 0.0)
                self.assertEqual(
                    link_res["status"].at[t, "pipe2"],
                    self.wntr.network.LinkStatus.Closed,
                )


class TestConditionalControls(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

    @classmethod
    def tearDownClass(self):
        pass

    
    def test_close_link_by_tank_level(self):
        inp_file = join(test_datadir, "conditional_controls_1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.report_timestep = "all"
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0

        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        activated_flag = False
        count = 0
        node_res = results.node
        link_res = results.link
        for t in results.time:
            if node_res["pressure"].at[t, "tank1"] >= 50.0 and not activated_flag:
                activated_flag = True
            if activated_flag:
                self.assertAlmostEqual(link_res["flowrate"].at[t, "pump1"], 0.0)
                self.assertEqual(
                    link_res["status"].at[t, "pump1"],
                    self.wntr.network.LinkStatus.Closed,
                )
                count += 1
            else:
                self.assertGreaterEqual(link_res["flowrate"].at[t, "pump1"], 0.0001)
                self.assertEqual(
                    link_res["status"].loc[t, "pump1"],
                    self.wntr.network.LinkStatus.Open,
                )
        self.assertEqual(activated_flag, True)
        self.assertGreaterEqual(count, 2)

    def test_update_priority(self):
        inp_file = join(test_datadir, "conditional_controls_1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        new_priority =1 
        idx = 0
        # check current priority is different to the new one
        self.assertNotEqual(wn.get_control(wn.control_name_list[idx]).priority, new_priority)
        # Update priority and check it has worked
        wn.get_control(wn.control_name_list[idx]).update_priority(new_priority)
        self.assertEqual(wn.get_control(wn.control_name_list[idx]).priority, new_priority)

    def test_update_conditions(self):
        inp_file = join(test_datadir, "conditional_controls_1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        idx = 0
        new_condition = self.wntr.network.controls.TimeOfDayCondition(wn, 'Is','01:00:00')

        self.assertNotEqual(wn.get_control(wn.control_name_list[idx]).condition, new_condition)
        wn.get_control(wn.control_name_list[idx]).update_condition(new_condition)
        self.assertEqual(wn.get_control(wn.control_name_list[idx]).condition, new_condition)


    def test_update_then_actions(self):
        inp_file = join(test_datadir, "conditional_controls_1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        link_num=0
        link = wn.get_link(wn.link_name_list[link_num])

        new_action = self.wntr.network.controls.ControlAction(link, 'status', 0)
        # When updating then_actions, the action will be made iterable, we need to make it iterable too.
        iterable_action = self.wntr.network.controls._ensure_iterable(new_action)

        self.assertNotEqual(wn.get_control(wn.control_name_list[0])._then_actions,iterable_action)
        wn.get_control(wn.control_name_list[0]).update_then_actions(new_action)
        self.assertEqual(wn.get_control(wn.control_name_list[0])._then_actions,iterable_action)

    def test_update_else_actions(self):
        inp_file = join(test_datadir, "conditional_controls_1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        link_num=0
        link = wn.get_link(wn.link_name_list[link_num])

        new_action = self.wntr.network.controls.ControlAction(link, 'status', 0)
        # When updating then_actions, the action will be made iterable, we need to make it iterable too.
        iterable_action = self.wntr.network.controls._ensure_iterable(new_action)

        self.assertNotEqual(wn.get_control(wn.control_name_list[0])._else_actions,iterable_action)
        wn.get_control(wn.control_name_list[0]).update_else_actions(new_action)
        self.assertEqual(wn.get_control(wn.control_name_list[0])._else_actions,iterable_action)

    def test_open_link_by_tank_level(self):
        inp_file = join(test_datadir, "conditional_controls_2.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.report_timestep = "all"
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        activated_flag = False
        count = 0  # Used to make sure the link is opened for at least 2 timesteps to make sure the link stays open
        for t in results.node["pressure"].index:
            if results.node["pressure"].at[t, "tank1"] >= 300.0 and not activated_flag:
                activated_flag = True
            if activated_flag:
                self.assertGreaterEqual(results.link["flowrate"].at[t, "pipe1"], 0.002)
                self.assertEqual(
                    results.link["status"].at[t, "pipe1"],
                    self.wntr.network.LinkStatus.Open,
                )
                count += 1
            else:
                self.assertAlmostEqual(results.link["flowrate"].at[t, "pipe1"], 0.0)
                self.assertEqual(
                    results.link["status"].at[t, "pipe1"],
                    self.wntr.network.LinkStatus.Closed,
                )
        self.assertEqual(activated_flag, True)
        self.assertGreaterEqual(count, 2)
        self.assertEqual(
            results.link["status"].at[results.link["status"].index[0], "pipe1"],
            self.wntr.network.LinkStatus.Closed,
        )  # make sure the pipe starts closed
        self.assertLessEqual(
            results.node["pressure"].at[results.node["pressure"].index[0], "tank1"],
            300.0,
        )  # make sure the pipe starts closed


class TestTankControls(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

    @classmethod
    def tearDownClass(self):
        pass

    def test_pipe_closed_for_low_level(self):
        inp_file = join(test_datadir, "tank_controls_1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        tank_level_dropped_flag = False
        for t in results.node["pressure"].index:
            if results.node["pressure"].at[t, "tank1"] <= 10.0:
                self.assertLessEqual(results.link["flowrate"].at[t, "pipe1"], 0.0)
                self.assertEqual(
                    results.link["status"].at[t, "pipe1"],
                    self.wntr.network.LinkStatus.Closed,
                )
                tank_level_dropped_flag = True
        self.assertEqual(tank_level_dropped_flag, True)

    def test_reopen_pipe_after_tank_fills_back_up(self):
        """
        inp_file = join(test_datadir, 'tank_controls_2.inp')
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = 'PDA'
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        tank_level_dropped_flag = False
        tank_refilled_flag = False
        for t in results.time:
            if results.node.loc['pressure',t,'tank1'] <= 10.0:
                self.assertLessEqual(results.link.loc['flowrate',t,'pipe1'],0.0)
                tank_level_dropped_flag = True
            elif results.node.loc['pressure',t,'tank1'] > 10.0:
                self.assertGreaterEqual(results.link.loc['flowrate',t,'pipe1'],0.001)
                if tank_level_dropped_flag:
                    tank_refilled_flag = True
        self.assertEqual(tank_level_dropped_flag, True)
        self.assertEqual(tank_refilled_flag, True)
        """
        raise unittest.SkipTest
        self.assertEqual(True, False)


class TestValveControls(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

    @classmethod
    def tearDownClass(self):
        pass

    def test_check_valve_closed(self):
        inp_file = join(test_datadir, "cv_controls.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        flowrate = results.link["flowrate"]
        for t in flowrate.index:
            self.assertAlmostEqual(flowrate.at[t, "pipe1"], 0.0)

    def test_check_valve_opened(self):
        inp_file = join(test_datadir, "cv_controls.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        tank1 = wn.get_node("tank1")
        tank2 = wn.get_node("tank2")
        tank1_init_level = tank1.init_level
        tank1.init_level = tank2.init_level
        tank2.init_level = tank1_init_level
        tank1._head = tank1.init_level + tank1.elevation
        tank2._head = tank2.init_level + tank2.elevation
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        flag1 = False
        flag2 = False
        for t in results.node["head"].index:
            if (
                results.node["head"].at[t, "tank1"]
                >= results.node["head"].at[t, "tank2"]
            ):
                self.assertGreaterEqual(results.link["flowrate"].at[t, "pipe1"], 0.001)
                flag1 = True
            else:
                self.assertAlmostEqual(results.link["flowrate"].at[t, "pipe1"], 0.0)
                flag2 = True

        self.assertEqual(flag1, True)
        self.assertEqual(flag2, True)


class TestControlCombinations(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

    @classmethod
    def tearDownClass(self):
        pass

    def test_open_by_time_close_by_condition(self):
        inp_file = join(test_datadir, "control_comb.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        control_action = self.wntr.network.ControlAction(
            wn.get_link("pipe1"), "status", self.wntr.network.LinkStatus.Opened
        )
        control = self.wntr.network.controls.Control._time_control(
            wn, 6 * 3600, "SIM_TIME", False, control_action
        )
        wn.add_control("open_time_6", control)
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        flag1 = False
        flag2 = False
        for t in results.node["head"].index:
            if t == 6 * 3600:
                flag1 = True
            if (
                t > 0
                and (
                    results.node["head"].at[t - 3600, "tank1"]
                    + (
                        results.node["demand"].at[t - 3600, "tank1"]
                        * 3600
                        * 4
                        / (3.14159 * wn.get_node("tank1").diameter ** 2)
                    )
                )
                <= 30
            ):
                flag1 = False
                flag2 = True
            if flag1 == False:
                self.assertAlmostEqual(results.link["flowrate"].at[t, "pipe1"], 0.0)
            elif flag1 == True:
                self.assertGreaterEqual(results.link["flowrate"].at[t, "pipe1"], 0.001)

        self.assertEqual(flag1, False)
        self.assertEqual(flag2, True)

    def test_close_by_condition_open_by_time_stay(self):
        inp_file = join(test_datadir, "control_comb.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        tank1 = wn.get_node("tank1")
        tank1.init_level = 40.0
        tank1._head = tank1.elevation + 40.0
        pipe1 = wn.get_link("pipe1")
        pipe1._user_status = self.wntr.network.LinkStatus.Opened
        control_action = self.wntr.network.ControlAction(
            wn.get_link("pipe1"), "status", self.wntr.network.LinkStatus.Opened
        )
        control = self.wntr.network.controls.Control._time_control(
            wn, 19 * 3600, "SIM_TIME", False, control_action
        )
        wn.add_control("open_time_19", control)
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        flag1 = False
        flag2 = False
        for t in results.node["head"].index:
            if t == 19 * 3600:
                flag1 = False
            if (
                t > 0
                and (
                    results.node["head"].at[t - 3600, "tank1"]
                    + (
                        results.node["demand"].at[t - 3600, "tank1"]
                        * 3600
                        * 4
                        / (3.14159 * wn.get_node("tank1").diameter ** 2)
                    )
                )
                <= 30
            ):
                flag1 = True
                flag2 = True
            if flag1 == False:
                self.assertGreaterEqual(results.link["flowrate"].at[t, "pipe1"], 0.001)
            elif flag1 == True:
                self.assertAlmostEqual(results.link["flowrate"].at[t, "pipe1"], 0.0)

        self.assertEqual(flag1, False)
        self.assertEqual(flag2, True)

    def test_close_by_condition_open_by_time_reclose(self):
        inp_file = join(test_datadir, "control_comb.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        tank1 = wn.get_node("tank1")
        tank1.init_level = 40.0
        tank1._head = tank1.elevation + 40.0
        pipe1 = wn.get_link("pipe1")
        pipe1._user_status = self.wntr.network.LinkStatus.Opened
        control_action = self.wntr.network.ControlAction(
            wn.get_link("pipe1"), "status", self.wntr.network.LinkStatus.Opened
        )
        control = self.wntr.network.controls.Control._time_control(
            wn, 5 * 3600, "SIM_TIME", False, control_action
        )
        wn.add_control("open_time_5", control)
        for jname, j in wn.nodes(self.wntr.network.Junction):
            j.minimum_pressure = 0.0
            j.required_pressure = 15.0
        wn.options.hydraulic.demand_model = "PDA"
        sim = self.wntr.sim.WNTRSimulator(wn)
        results = sim.run_sim()

        flag1 = False
        for t in results.node["head"].index:
            if (
                t > 0
                and (
                    results.node["head"].at[t - 3600, "tank1"]
                    + (
                        results.node["demand"].at[t - 3600, "tank1"]
                        * 3600
                        * 4
                        / (3.14159 * wn.get_node("tank1").diameter ** 2)
                    )
                )
                <= 30.0
            ):
                flag1 = True
            if t == 5 * 3600:
                flag1 = False
            if flag1 == False:
                self.assertGreaterEqual(results.link["flowrate"].at[t, "pipe1"], 0.001)
            elif flag1 == True:
                self.assertAlmostEqual(results.link["flowrate"].at[t, "pipe1"], 0.0)

        self.assertEqual(flag1, True)
        
class TestControlParsing(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr
        import numpy as np
        self.wntr = wntr
        self.np = np
        
        # Create a simple water network for testing
        self.wn = wntr.network.WaterNetworkModel()
        inp_file = join(test_datadir, "control_comb.inp")
        self.wn = wntr.network.WaterNetworkModel(inp_file)
        self.flow_units = wntr.epanet.util.FlowUnits['SI']

        
    @classmethod
    def tearDownClass(self):
        pass
    
    def test_controls_format_link_status_time(self):
        """Test CONTROLS format: LINK element status AT TIME"""
        line = "LINK pipe1 OPEN AT TIME 12"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a time control (presolve type)
        self.assertEqual(control.epanet_control_type.name, "presolve")
        # Check the condition is SimTimeCondition
        self.assertIsInstance(control._condition, self.wntr.network.controls.SimTimeCondition)
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Opened)
        # Check the time threshold
        self.assertEqual(control._condition._threshold, 12 * 3600)  # 12 hours in seconds
        
    def test_controls_format_link_status_clocktime(self):
        """Test CONTROLS format: LINK element status AT CLOCKTIME"""
        line = "LINK pipe1 CLOSED AT CLOCKTIME 8:30 AM"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a time control (presolve type)
        self.assertEqual(control.epanet_control_type.name, "presolve")
        # Check the condition is TimeOfDayCondition
        self.assertIsInstance(control._condition, self.wntr.network.controls.TimeOfDayCondition)
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Closed)
        
    def test_controls_format_link_status_conditional(self):
        """Test CONTROLS format: LINK element status IF NODE condition"""
        line = "LINK pipe1 CLOSED IF NODE tank1 BELOW 30"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a conditional control (pre_and_postsolve type for tank levels)
        self.assertEqual(control.epanet_control_type.name, "pre_and_postsolve")
        # Check the condition is TankLevelCondition (ValueCondition subclass)
        self.assertIsInstance(control._condition, self.wntr.network.controls.TankLevelCondition)
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Closed)
        # Check the condition details
        self.assertEqual(control._condition._source_obj.name, "tank1")
        self.assertEqual(control._condition._source_attr, "level")
        
    def test_controls_format_pump_setting(self):
        """Test CONTROLS format: LINK pump setting"""
        line = "LINK pump1 1.5 AT TIME 5"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a time control (presolve type)
        self.assertEqual(control.epanet_control_type.name, "presolve")
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pump1")
        self.assertEqual(action._attribute, "base_speed")
        self.assertEqual(action._value, 1.5)
        
    def test_rules_format_time_condition(self):
        """Test RULES format: IF SYSTEM TIME THEN action"""
        line = "IF SYSTEM TIME IS 6 THEN LINK pipe1 IS OPEN"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a time control (presolve type)
        self.assertEqual(control.epanet_control_type.name, "presolve")
        # Check the condition is SimTimeCondition
        self.assertIsInstance(control._condition, self.wntr.network.controls.SimTimeCondition)
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Opened)
        # Check the time threshold
        self.assertEqual(control._condition._threshold, 6 * 3600)  # 6 hours in seconds
        
    def test_rules_format_clocktime_condition(self):
        """Test RULES format: IF SYSTEM CLOCKTIME THEN action"""
        line = "IF SYSTEM CLOCKTIME IS 9:00 AM THEN LINK pipe1 IS CLOSED"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a time control (presolve type)
        self.assertEqual(control.epanet_control_type.name, "presolve")
        # Check the condition is TimeOfDayCondition
        self.assertIsInstance(control._condition, self.wntr.network.controls.TimeOfDayCondition)
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Closed)
        
    def test_rules_format_tank_level_condition(self):
        """Test RULES format: IF TANK level condition THEN action"""
        line = "IF TANK tank1 LEVEL BELOW 25 THEN LINK pump1 IS OPEN"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a conditional control (pre_and_postsolve type for tank levels)
        self.assertEqual(control.epanet_control_type.name, "pre_and_postsolve")
        # Check the condition is TankLevelCondition
        self.assertIsInstance(control._condition, self.wntr.network.controls.TankLevelCondition)
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pump1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Opened)
        # Check the condition details
        self.assertEqual(control._condition._source_obj.name, "tank1")
        self.assertEqual(control._condition._source_attr, "level")
        
    def test_rules_format_junction_pressure_condition(self):
        """Test RULES format: IF JUNCTION pressure condition THEN action"""
        line = "IF JUNCTION j1 PRESSURE ABOVE 20 THEN LINK pipe1 IS CLOSED"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a conditional control (postsolve type for pressure conditions)
        self.assertEqual(control.epanet_control_type.name, "postsolve")
        # Check the condition is ValueCondition
        self.assertIsInstance(control._condition, self.wntr.network.controls.ValueCondition)
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Closed)
        # Check the condition details
        self.assertEqual(control._condition._source_obj.name, "j1")
        self.assertEqual(control._condition._source_attr, "pressure")
        
    def test_empty_line(self):
        """Test empty line handling"""
        line = ""
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        self.assertIsNone(control)
        
    def test_comment_line(self):
        """Test line with comment only"""
        line = "; This is a comment"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        self.assertIsNone(control)
        
    def test_line_with_comment(self):
        """Test line with control and comment"""
        line = "LINK pipe1 OPEN AT TIME 12 ; This is a comment"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a time control (presolve type)
        self.assertEqual(control.epanet_control_type.name, "presolve")
        # Check the action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1")
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Opened)
        
    def test_time_format_hours_minutes(self):
        """Test time format with hours:minutes"""
        line = "LINK pipe1 OPEN AT TIME 12:30"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check that it's a time control (presolve type)
        self.assertEqual(control.epanet_control_type.name, "presolve")
        expected_seconds = 12 * 3600 + 30 * 60  # 12 hours 30 minutes in seconds
        self.assertEqual(control._condition._threshold, expected_seconds)
        
    def test_node_status_control(self):
        """Test NODE element status control"""
        line = "NODE j1 TRUE AT TIME 5"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "j1")
        self.assertEqual(action._attribute, "leak_status")
        self.assertEqual(action._value, True)
        
        # Check condition
        self.assertEqual(control._condition._threshold, 5 * 3600)

        
    def test_invalid_rule(self):
        """Test invalid RULES format"""
        line = "IF SYSTEM TIME IS 6 LINK pipe1 IS OPEN"
        with self.assertRaises(RuntimeError):
            _read_control_line(line, self.wn, self.flow_units, "test_control")
            
    def test_operator_aliases(self):
        """Test different operator aliases (ABOVE vs >, BELOW vs <)"""
        line1 = "LINK pipe1 CLOSED IF NODE tank1 ABOVE 30"
        line2 = "LINK pipe1 CLOSED IF NODE tank1 > 30"
        
        control1 = _read_control_line(line1, self.wn, self.flow_units, "test_control1")
        control2 = _read_control_line(line2, self.wn, self.flow_units, "test_control2")
        
        # Both should create equivalent controls
        self.assertEqual(control1.epanet_control_type.name, control2.epanet_control_type.name)
        self.assertEqual(control1._condition._source_obj.name, control2._condition._source_obj.name)
        self.assertEqual(control1._condition._source_attr, control2._condition._source_attr)

    def test_existing_control_from_file(self):
        """Test that the existing control from control_comb.inp is parsed correctly"""
        # The file contains: LINK pipe1 CLOSED IF NODE tank1 BELOW 30
        line = "LINK pipe1 CLOSED IF NODE tank1 BELOW 30"
        control = _read_control_line(line, self.wn, self.flow_units, "test_control")
        
        # Check action
        action = control._then_actions[0]
        self.assertEqual(action._target_obj.name, "pipe1") 
        self.assertEqual(action._attribute, "status")
        self.assertEqual(action._value, self.wntr.network.LinkStatus.Closed)
        # Check condition
        self.assertEqual(control._condition._source_obj.name, "tank1")
        self.assertEqual(control._condition._source_attr, "level")
        self.assertEqual(control._condition._relation, self.wntr.network.controls.Comparison.lt)


if __name__ == "__main__":
    unittest.main()
